# from genet.utils import *
import os, sys, regex
import pandas as pd
from Bio import Entrez, GenBank
from Bio.Seq import reverse_complement, translate
from Bio.SeqUtils import gc_fraction
from genet.design.DesignUtils import dict_pam_disrup_rank, test_score_data

'''
TODO
1. flash를 python code로 구현한 것이 없으므로, 여기서 input은 .fastq 파일만 가능
2. 나중에 flashpy 개발이 좀 더 진행되면 도입하는 것을 생각해보자
3. python 기본적으로 python 3.6~3.10까지 호환되는 것을 목표로 하고, 3.11도 테스트하기

'''


class MakeStop:
    '''
    특정 gene symbol을 넣으면,
    그 gene에서 SNV (1bp substitution)으로 만들 수 있는 모든 stop codon의 경우의 수를 return 한다.

    input  = gene symbol (ensembl 기준)
    output = DataFrame (gene | pos | WT (121bp) | ED (121bp))

    example:
    >>> from genet.design import MakeStop
    >>> df_out = MakeStop('ACE2')

    '''
    def __init__(self, gene_name:str):
        print('Start MakeStop')

# class END: MakeStop



class MakeSNVs:

    def __init__(self, seq:str, pos:int) -> list:
        """sequence에서 지정된 위치에 A/T/G/C로 만들어진 SNV를 만들어서 list로 return하는 함수

        Args:
            * seq (str): SNV를 만들 서열 정보
            * pos (int): seq에서 SNV를 만들 위치 정보

        Returns:
            list: 특정 위치에 가능한 SNV들이 모인 list

        Eaxmple
            >>> from genet.design import MakeSNVs
            >>> seq = 'ATGGCTGACTGC'
            >>> list_snvs = MakeSNVs(seq, 4)

            list_snvs = ['ATGGATGACTGC', 'ATGGGTGACTGC', 'ATGGTTGACTGC']
        """        

        list_base = ['A', 'G', 'C', 'T']
        self.list_sSNV = []

        for base in list_base:
            list_sSeq = list(seq)
            list_sSeq[pos] = base
            snv_temp = ''.join(list_sSeq)
            if seq!=snv_temp: self.list_sSNV.append(snv_temp)

# class END: MakeSNVs


class SynonymousPE:

    def __init__(self,
                 dp_record:pd.Series,
                 ref_seq:str,
                 frame:int,
                 cds_start:int=0,
                 cds_end:int=121,
                 adj_rha:bool=True,
                 mut_target:str=None,
                 ):
        """DeepPrime output으로 만들어진 파일에서 pegRNA들에 silent mutation을 함께 유발하는 것 중 최적의 pegRNA를 선택해주는 것
        모든 기능은 prime editing으로 1bp substitution을 했을 때를 가정하여 만들어졌다.
        우선, intended edit 기준으로 best pegRNA들을 선정한다.
        그 후, 아래와 같이 silent PAM co-editing을 추가한다.

        Case 1: Silent PAM co-editing이 LHA 부분에 발생하는 경우,
            그대로 silent PAM co-editing을 넣고 완성

        Case 2: Silent PAM co-editing이 RHA 부분에 발생하는 경우,
            원래 pegRNA의 RHA 길이를 기준으로 맞춰주기 위해 RTT 길이를 늘려주기.
        
        Case 3: Intended edit이 이미 PAM 위치 (+5 or +6)인 경우,
        1) 만약 남은 PAM 위치 (G)에 silent mutation이 가능하면 그 것으로 결정
        2) 남은 PAM 위치의 SNV 중 silent mutation이 없다면, LHA에서 가능한 silent mutation을 넣어준다.
        3) 만약 선택되는 mutation position이 splicing adaptor 위치라면, (cds_start / end 앞뒤로 5nt) 차순위로 뽑기
        4) 차순위로 불가능한 것들이 있다면, RHA에서 silent mutation을 뽑는다.         
    
        Args:
            * dp_record (pd.Series): DeepPrime을 돌리고 output으로 나오는 것의 한 index의 정보를 가져온 것.
            * ref_seq (str): 해당 mutation을 만들었을 떄, 사용한 reference sequence (121nt)
            * frame (int): Ref_seq의 frame을 표시 (0, 1, 2)
            * cds_start (int, optional): CDS가 시작되는 위치, 그 이전 위치에서부터는 silent mutation을 만들 수 없으므로 고르는 위치에서 제외. Defaults to 0.
            * cds_end (int, optional): CDS가 종료되는 위치, 그 이후 위치에서부터는 silent mutation을 만들 수 없으므로 고르는 위치에서 제외. Defaults to 121.
            * adj_rha (bool, optional): silent mutation이 RHA에 위치하는 경우, 기존의 pegRNA에서의 RHA 길이만큼을 유지하기 위해 RTT를 늘려주는 기능. Defaults to True.
            * mut_target (str, optional): Synonymous mutation을 특정 패턴의 sequence에서 찾아서 하고 싶을 경우. 해당 sequence가 들어있는 codon에 만든다.

        Raises:
            ValueError: frame이 0, 1, 2 중에 하나로 입력되지 않은 경우
            ValueError: Reference sequence가 pegRNA와 matching이 안 될 경우
        """        
    
        # input error check
        if type(dp_record) != type(pd.Series()): raise TypeError("The type of 'dp_record' should be pd.Series.")
        if frame not in [0, 1, 2]              : raise ValueError('Frame should be 0, 1 or 2')

        # step 1: Get required features
        '''
        cds_start / end / frame 등을 variants마다 각각 따로 넣어줘야 하는 것이 복잡한 것 같음
        Exon의 길이 (fixed), exon의 frame, SNV의 position만 넣으면, 
        각 SNV의 frame, CDS start, CDS end를 내부적으로 만들어서 계산하는 것으로 수정하기

        예를 들어, cds_end는 다음과 같이 구할 수 있음
        cds_end = 66 + exon_len - pos 
        if cds_end > 121: cds_end=121
        
        '''

        self.rec        = dp_record
        self.sID        = dp_record.ID
        self.wt_seq     = dp_record.Target
        self.rtpbs_dna  = reverse_complement(dp_record['RT-PBS'])
        self.pbs_dna    = self.rtpbs_dna[:dp_record.PBS_len]
        self.rtt_dna    = self.rtpbs_dna[-dp_record.RTT_len:]

        self.rtt_len    = dp_record.RTT_len

        self.edit_pos   = dp_record.Edit_pos
        self.ref_seq    = ref_seq.upper()
        self.frame      = frame
        self.cds_start  = cds_start
        self.cds_end    = cds_end
        self.mut_target = mut_target.upper()
        
        self.splicing_adaptor = [i for i in range(cds_start-5, cds_start)] + [i for i in range(cds_end, cds_end+5)]

        # step 2: pegRNA의 strand 방향에 따라 synonymous Mut 생성 함수 결정
        if self.wt_seq in self.ref_seq:
            self.strand = '+'
            self.rtt_frame = (frame - self.edit_pos + 1) % 3 # rtt 시작점의 frame, LHA 길이를 이용한 계산

        elif reverse_complement(self.wt_seq) in self.ref_seq:
            self.strand = '-'
            self.rtt_frame = (self.edit_pos + frame) % 3  # revcom_rtt_dna의 3' end가 위치하는 지점의 frame. 시작점이 거기이기 때문.
            
        else: raise ValueError('Reference sequence is not matched with pegRNA information!\nPlease chech your ref_seq')
    
        self.dict_mut = {
            'Codon_WT'      : [],
            'Codon_Mut'     : [],
            'RTT_DNA_frame' : [],
            'RTT_DNA_Strand': [],
            'AminoAcid_WT'  : [],
            'AminoAcid_Mut' : [],
            'Silent_check'  : [],
            'Mut_pos'       : [],
            'Mut_refpos'    : [],
            'PAM_Mut'       : [],
            'Priority'      : [],
            'Edit_class'    : [],
            'RTT_DNA_Mut'   : [],
        }
        
        if self.mut_target==None:
            # edit position / Silent mut 고려해서 뽑은 것들
            self.output = self.make_synonyPAM_RTT(self.rtt_frame, self.rtt_dna, self.strand)
        
        else:
            if (self.mut_target in self.rtt_dna) or (reverse_complement(self.mut_target) in self.rtt_dna):
                self.output = self.make_synonySeqMut_RTT(self.rtt_frame, self.rtt_dna, self.strand, self.mut_target)
            else:
                self.output = self.make_synonyPAM_RTT(self.rtt_frame, self.rtt_dna, self.strand)
        
        # step 3: 만약 RHA 길이 조정 옵션이 True로 되어있으면, 조정해주기. (defualt)
        if adj_rha == True:
            adj_len = self.output['Mut_pos'] - self.edit_pos
            
            if adj_len > 0: 
                rtt_end = 21 + self.rtt_len
                self.output['RTT_DNA_Mut'] = self.output['RTT_DNA_Mut'] + self.wt_seq[rtt_end:rtt_end+adj_len]

        self.extension = self.pbs_dna + self.output['RTT_DNA_Mut']
        
    # End def __init__:

        
    def make_snv(self, seq:str, pos:int) -> list:
        """sequence에서 지정된 위치에 A/T/G/C로 만들어진 SNV를 만들어서 list로 return하는 함수

        Args:
            seq (str): SNV를 만들 서열 정보
            pos (int): seq에서 SNV를 만들 위치 정보

        Returns:
            list: 특정 위치에 가능한 SNV들이 모인 list
        """        

        list_base = ['A', 'G', 'C', 'T']
        list_sSNV = []

        for base in list_base:
            list_sSeq = list(seq)
            list_sSeq[pos] = base
            snv_temp = ''.join(list_sSeq)
            if seq!=snv_temp: list_sSNV.append(snv_temp)

        return list_sSNV
    # def END: make_snv

        
    def make_synonySeqMut_RTT(self, rtt_frame:int, rtt_dna:str, strand:str, mut_target:str) -> pd.DataFrame:
        """특정 sequence motif가 주어진다면, 그 sequence가 RTT에 있는지 확인해본다.
        만약 RTT 안에 해당 sequence가 있으면 그 sequence 안에 silent mutation을 만드는 것을 최우선으로 한다. 

        Args:
            * rtt_frame (int): CDS에서 RTT의 frame을 의미함 (0, 1, 2).
            * rtt_dna (str): pegRNA의 RTT 부분을 DNA로 가져온 것. cDNA와 동일.
            * strand (str): Reference sequence 기준으로 pegRNA의 방향 (+ / -)
            * mut_target (str): synonymous mutation을 만들 target 서열 (ex: restriction enzyme cut site)

        Returns:
            pd.DataFrame: PAM 위치에 가능한 synonymous mutation 정보들
        """    
        ep = self.edit_pos
        mt = mut_target

        if mt in rtt_dna: mt_dna = mt
        else            : mt_dna = reverse_complement(mt)
        
        ext_rtt_dna = self.wt_seq[:21] + rtt_dna + self.wt_seq[21+len(rtt_dna):] # WT sequence before nick pos.
        mt_pos      = rtt_dna.find(mt_dna)
        mt_start    = 21 + mt_pos
        mt_end      = 21 + mt_pos + len(mt)

        if strand == '+':
            mt_frame = (rtt_frame + mt_pos) % 3
            codon_le = mt_frame
            codon_re = (3 - (mt_frame + len(mt))) % 3
            codon_mt = ext_rtt_dna[mt_start-codon_le:mt_start] + mt_dna + ext_rtt_dna[mt_end:mt_end+codon_re]

        else:
            mt_frame = (rtt_frame - mt_pos) % 3
            codon_le = (mt_frame - len(mt)) % 3
            codon_re = (mt_pos - rtt_frame) % 3
            codon_mt =  reverse_complement(ext_rtt_dna[mt_start-codon_re:mt_start] + mt_dna + ext_rtt_dna[mt_end:mt_end+codon_le])

        dict_codon_mtPos = {codon_mt: [i for i in range(codon_le, len(codon_mt) - codon_re)]}

        self.dict_codon_mtPos = dict_codon_mtPos

        # for loop: LHA 위치가 걸쳐있는 codon들을 가져온다.
        for codon in dict_codon_mtPos:
            # for loop: 각 codon들에서 mut_target에 해당하는 위치들을 가져온다.
            for snv_pos in dict_codon_mtPos[codon]:
                
                if strand == '+': mut_pos = mt_pos - (codon_le - snv_pos) + 1
                else            : mut_pos = mt_pos + (codon_le - snv_pos) + len(mt_dna)

                if mut_pos == ep: continue
                if mut_pos > self.rtt_len: continue

                list_mut_codon = self.make_snv(codon, snv_pos)
                
                for mut_codon in list_mut_codon:
                    
                    aa_wt  = translate(codon)
                    aa_mut = translate(mut_codon)
                    
                    if mut_pos in [5, 6]: priority = 0
                    else:
                        priority = abs(mut_pos - ep)
                        if gc_fraction(codon) != gc_fraction(mut_codon): priority += 1

                    if strand == '+':
                        mut_refpos  = 60 + (mut_pos - ep)
                        rtt_dna_mut = self.rtt_dna[:mut_pos-1] + mut_codon[snv_pos]+ self.rtt_dna[mut_pos:]

                    else:
                        mut_refpos  = 60 - (mut_pos - ep)
                        rtt_dna_mut = self.rtt_dna[:mut_pos-1] + reverse_complement(mut_codon[snv_pos]) + self.rtt_dna[mut_pos:]

                    if mut_refpos in self.splicing_adaptor: silent_check = False
                    else                                  : silent_check = aa_wt==aa_mut

                    self.dict_mut['Codon_WT'].append(codon)
                    self.dict_mut['Codon_Mut'].append(mut_codon)
                    self.dict_mut['RTT_DNA_frame'].append(rtt_frame)
                    self.dict_mut['RTT_DNA_Strand'].append(strand)
                    self.dict_mut['AminoAcid_WT'].append(aa_wt)
                    self.dict_mut['AminoAcid_Mut'].append(aa_mut)
                    self.dict_mut['Silent_check'].append(silent_check)
                    self.dict_mut['Mut_pos'].append(mut_pos)
                    self.dict_mut['Mut_refpos'].append(mut_refpos)
                    self.dict_mut['PAM_Mut'].append(rtt_dna_mut[4:6])
                    self.dict_mut['Priority'].append(priority)
                    self.dict_mut['Edit_class'].append('MutSeq_edit')
                    self.dict_mut['RTT_DNA_Mut'].append(rtt_dna_mut)

        self.mutations  = pd.DataFrame(self.dict_mut) 

        df_synonymous = self.mutations.groupby(by='Silent_check').get_group(True).sort_values(by='Priority').reset_index(drop=True)

        return df_synonymous.iloc[0]

    # def END: make_synonymousSeq_RTT


    def make_dict_codon_pamPos(self, strand:str, rtt_frame:int, rtt_dna:str) -> dict:
        """_summary_

        Args:
            strand (str): _description_
            rtt_frame (int): _description_
            rtt_dna (str): _description_

        Returns:
            dict: _description_
        """       


        rtt_end = 21 + self.rtt_len

        if strand == '+':
            ext_rtt_dna = rtt_dna+self.wt_seq[rtt_end:]
            if   rtt_frame == 0: dict_codon_pamPos = {(ext_rtt_dna)[3:6]: [1, 2]}
            elif rtt_frame == 1: dict_codon_pamPos = {(ext_rtt_dna)[2:5]: [2]}
            else               : dict_codon_pamPos = {(ext_rtt_dna)[4:7]: [0, 1]}

        
        else:
            # strand가 (-)인 경우에는 RT-PBS를 revcom으로 바꿔서 PAM 기준으로 SNV들을 만들기
            ext_rtt_dna = reverse_complement(rtt_dna+self.wt_seq[rtt_end:])
            if   rtt_frame == 0: dict_codon_pamPos = {ext_rtt_dna[-6:-3]: [0, 1]}
            elif rtt_frame == 1: dict_codon_pamPos = {ext_rtt_dna[-7:-4]: [1, 2]}
            else               : dict_codon_pamPos = {ext_rtt_dna[-8:-5]: [2], ext_rtt_dna[-5:-2]: [0]}

        return dict_codon_pamPos


    
    def make_synonyPAM_RTT(self, rtt_frame:int, rtt_dna:str, strand:str) -> pd.DataFrame:
        """PAM sequence (GG)에서 frame에 따라 가능한 synonymous mutation들을 만들고,
        이에 대한 결과를 DataFrame으로 만들어준다. 

        Args:
            * rtt_frame (int): CDS에서 RTT의 frame을 의미함 (0, 1, 2).
            * rtt_dna (str): pegRNA의 RTT 부분을 DNA로 가져온 것. cDNA와 동일.
            * strand (str): Reference sequence 기준으로 pegRNA의 방향 (+ / -)

        Returns:
            pd.DataFrame: PAM 위치에 가능한 synonymous mutation 정보들
        """   
        
        ep = self.edit_pos

        dict_codon_pamPos = self.make_dict_codon_pamPos(strand, rtt_frame, rtt_dna)


        if strand == '+': PAM_G_pos = 5
        else            : PAM_G_pos = 6
        
        try:
            # for loop: GG PAM 위치가 걸쳐있는 codon들을 가져온다.
            for codon in dict_codon_pamPos:
                # for loop: 각 codon들에서 GG PAM에 해당하는 위치들을 가져온다.
                for snv_pos in dict_codon_pamPos[codon]:
                    if PAM_G_pos > self.rtt_len:
                        if strand == '+': PAM_G_pos += 1
                        else            : PAM_G_pos -= 1
                        continue

                    list_mut_codon = MakeSNVs(codon, snv_pos).list_sSNV
                    
                    for mut_codon in list_mut_codon:

                        aa_wt  = translate(codon)
                        aa_mut = translate(mut_codon)
                        
                        rtt_end = 21 + self.rtt_len
                        ext_rtt_dna = rtt_dna+self.wt_seq[rtt_end:]

                        if strand == '+':
                            if PAM_G_pos == 5: pam_mut = mut_codon[snv_pos] + ext_rtt_dna[5]
                            else             : pam_mut = ext_rtt_dna[4] + mut_codon[snv_pos]
                            rtt_dna_mut = self.rtt_dna[:PAM_G_pos-1] + mut_codon[snv_pos] + self.rtt_dna[PAM_G_pos:]
                            mut_refpos  = 60 + (PAM_G_pos - ep)
                        
                        else:
                            if PAM_G_pos == 6: pam_mut = ext_rtt_dna[4] + reverse_complement(mut_codon[snv_pos])
                            else             : pam_mut = reverse_complement(mut_codon[snv_pos]) + ext_rtt_dna[5]
                            rtt_dna_mut = self.rtt_dna[:PAM_G_pos-1] + reverse_complement(mut_codon[snv_pos]) + self.rtt_dna[PAM_G_pos:]
                            mut_refpos  = 60 - (PAM_G_pos - ep)
                        
                        if mut_refpos in self.splicing_adaptor: silent_check = False
                        else                                  : silent_check = aa_wt==aa_mut

                        self.dict_mut['Codon_WT'].append(codon)
                        self.dict_mut['Codon_Mut'].append(mut_codon)
                        self.dict_mut['RTT_DNA_frame'].append(rtt_frame)
                        self.dict_mut['RTT_DNA_Strand'].append(strand)
                        self.dict_mut['AminoAcid_WT'].append(aa_wt)
                        self.dict_mut['AminoAcid_Mut'].append(aa_mut)
                        self.dict_mut['Silent_check'].append(silent_check)
                        self.dict_mut['Mut_pos'].append(PAM_G_pos)
                        self.dict_mut['Mut_refpos'].append(mut_refpos)
                        self.dict_mut['PAM_Mut'].append(pam_mut)
                        self.dict_mut['Priority'].append(dict_pam_disrup_rank[pam_mut])
                        self.dict_mut['RTT_DNA_Mut'].append(rtt_dna_mut)
                        self.dict_mut['Edit_class'].append('PAM_edit')

                    
                    if strand == '+': PAM_G_pos += 1
                    else            : PAM_G_pos -= 1
                        
            self.mutations  = pd.DataFrame(self.dict_mut) 

            df_synonymous = self.mutations.groupby(by='Silent_check').get_group(True).sort_values(by='Priority').reset_index(drop=True)
            df_synonymous = df_synonymous[df_synonymous['Mut_pos'] != self.edit_pos]

            return df_synonymous.iloc[0]

        except:
            if self.edit_pos > 1: df_synonymous = self.make_synonyLHA(self.rtt_dna, rtt_frame, strand)
            else                : df_synonymous = self.make_synonyRHA(self.rtt_dna, rtt_frame, strand)
            
            return df_synonymous.iloc[0]

    # def END: make_synonymousPAM_RTT

        
    def make_synonyLHA(self, rtt_dna:str, rtt_frame:int, strand:str) -> pd.DataFrame:
        """만약 PAM synonymous mutation이 불가능한 경우,
        silent mutation은 LHA에 만들어주는 함수

        Returns:
            pd.DataFrame: df_synonymous
        """

        ep = self.edit_pos

        if strand == '+':
            codon_le  = rtt_frame
            codon_re  = (3 - (rtt_frame + ep-1)) % 3
            codon_LHA = rtt_dna[:ep-1] + rtt_dna[ep-1:ep-1+codon_re]
            if codon_le > 0: codon_LHA = self.pbs_dna[-codon_le:] + codon_LHA
            dict_codon_LHAPos = {codon_LHA: [i for i in range(0, ep-1+codon_re)]}

        else:
            codon_le  = (3 - (ep - 1) + rtt_frame) % 3 # ? 이게 문제인듯?
            codon_re  = (3 - rtt_frame) % 3
            codon_LHA = rtt_dna[:ep-1] + rtt_dna[ep-1:ep-1+codon_le]
            if codon_re > 0: codon_LHA = self.pbs_dna[-codon_re:] + codon_LHA
            codon_LHA = reverse_complement(codon_LHA)
            dict_codon_LHAPos = {codon_LHA: [i for i in range(codon_le, codon_le+ep-1)]}

        self.codon_le = codon_le
        self.codon_re = codon_re
        self.codon_LHA = codon_LHA

        # for loop: LHA 위치가 걸쳐있는 codon들을 가져온다.
        for codon in dict_codon_LHAPos:
            # for loop: 각 codon들에서 LHA에 해당하는 위치들을 가져온다.
            for snv_pos in dict_codon_LHAPos[codon]:
                
                if strand == '+': mut_pos = snv_pos + 1 - rtt_frame
                else            : mut_pos = ep - snv_pos

                if mut_pos >= ep: continue
                if mut_pos <  1 : continue

                list_mut_codon = self.make_snv(codon, snv_pos)
                
                for mut_codon in list_mut_codon:
                    
                    aa_wt  = translate(codon)
                    aa_mut = translate(mut_codon)

                    if codon_re == 0: mut_LHA = mut_codon[codon_le:]
                    else            : mut_LHA = mut_codon[codon_le:-codon_re]
                    
                    priority = ep - mut_pos
                    if gc_fraction(codon) != gc_fraction(mut_codon): priority += 1
                    
                    if strand == '+':
                        rtt_dna_mut = mut_LHA + rtt_dna[ep-1:]
                        mut_refpos  = 60 + (mut_pos - ep)
                        
                    else:
                        rtt_dna_mut = reverse_complement(mut_LHA) + rtt_dna[ep-1:]
                        mut_refpos  = 60 - (mut_pos - ep)

                    if mut_refpos in self.splicing_adaptor: silent_check = False
                    else                                  : silent_check = aa_wt==aa_mut
                    
                    self.dict_mut['Codon_WT'].append(codon)
                    self.dict_mut['Codon_Mut'].append(mut_codon)
                    self.dict_mut['RTT_DNA_frame'].append(rtt_frame)
                    self.dict_mut['RTT_DNA_Strand'].append(strand)
                    self.dict_mut['AminoAcid_WT'].append(aa_wt)
                    self.dict_mut['AminoAcid_Mut'].append(aa_mut)
                    self.dict_mut['Silent_check'].append(silent_check)
                    self.dict_mut['Mut_pos'].append(mut_pos)
                    self.dict_mut['Mut_refpos'].append(mut_refpos)
                    self.dict_mut['Priority'].append(priority) # intended edit (PAM) 위치에 가까울수록 우선
                    self.dict_mut['PAM_Mut'].append(rtt_dna_mut[4:6])
                    self.dict_mut['RTT_DNA_Mut'].append(rtt_dna_mut)
                    self.dict_mut['Edit_class'].append('LHA_edit')

        self.mutations  = pd.DataFrame(self.dict_mut) 

        try:
            df_synonymous = self.mutations.groupby(by='Edit_class').get_group('LHA_edit').reset_index(drop=True)
            df_synonymous = df_synonymous.groupby(by='Silent_check').get_group(True).sort_values(by='Priority').reset_index(drop=True)
            return df_synonymous

        except:
            df_synonymous = self.make_synonyRHA(rtt_dna, rtt_frame, strand)

            return df_synonymous
    
    # def End: make_synonyLHA


    def make_synonyRHA(self, rtt_dna:str, rtt_frame:int, strand:str) -> pd.DataFrame:
        """만약 LHA synonymous mutation이 불가능한 경우,
        silent mutation은 RHA에 만들어주는 함수
        최후의 방법이다...

        Returns:
            pd.DataFrame: df_synonymous
        """
        
        ep = self.edit_pos
        rtt_end   = 21 + self.rtt_len
        ext_dna   = self.wt_seq[:21] + rtt_dna # WT sequence before nick pos.

        if strand == '+':
            RHA_frame = (rtt_frame + ep) % 3
            codon_le  = RHA_frame
            codon_re  = (3 - (RHA_frame + self.rtt_len - ep)) % 3 # (+) strand인 경우, codon 기준으로 RTT -> RTT end 방향이 RE 방향
            codon_RHA = ext_dna[21+ep-codon_le:21+ep]+ rtt_dna[ep:] + self.wt_seq[rtt_end:rtt_end+codon_re]

        else:
            RHA_frame = rtt_frame
            codon_le  = (3 - (self.rtt_len - rtt_frame)) % 3
            codon_re  = (ep - rtt_frame) % 3 # (-) strand인 경우, codon 기준으로 RTT -> Nick 방향이 RE 방향
            codon_RHA = ext_dna[22+ep-codon_re:22+ep]+ rtt_dna[ep:] + self.wt_seq[rtt_end:rtt_end+codon_le]
            codon_RHA = reverse_complement(codon_RHA)

        self.codon_le = codon_le
        self.codon_re = codon_re
        self.codon_RHA = codon_RHA

        dict_codon_RHAPos = {codon_RHA: [i for i in range(codon_le, len(codon_RHA))]}

        # for loop: LHA 위치가 걸쳐있는 codon들을 가져온다.
        for codon in dict_codon_RHAPos:
            # for loop: 각 codon들에서 LHA에 해당하는 위치들을 가져온다.
            for snv_pos in dict_codon_RHAPos[codon]:
                
                if strand == '+': mut_pos = ep + snv_pos - codon_le + 1
                else            : mut_pos = self.rtt_len + codon_le - snv_pos

                if mut_pos <= ep: continue
                if mut_pos > self.rtt_len: continue

                list_mut_codon = self.make_snv(codon, snv_pos)
                
                for mut_codon in list_mut_codon:
                    
                    aa_wt  = translate(codon)
                    aa_mut = translate(mut_codon)
                    
                    priority = mut_pos - ep
                    if gc_fraction(codon) != gc_fraction(mut_codon): priority += 1

                    self.dict_mut['Priority'].append(priority) # intended edit (PAM) 위치에 가까울수록 우선
                    
                    if strand == '+':
                        if codon_re == 0: mut_RHA = mut_codon[codon_le:]
                        else            : mut_RHA = mut_codon[codon_le:-codon_re]
                        mut_refpos  = 60 + (mut_pos - ep)

                    else:
                        if codon_re == 0: mut_RHA = reverse_complement(mut_codon[codon_le:])
                        else            : mut_RHA = reverse_complement(mut_codon[codon_le:-codon_re])
                        mut_refpos  = 60 - (mut_pos - ep)

                    rtt_dna_mut = self.rtt_dna[:ep] + mut_RHA

                    if mut_refpos in self.splicing_adaptor: silent_check = False
                    else                                  : silent_check = aa_wt==aa_mut

                    self.dict_mut['Codon_WT'].append(codon)
                    self.dict_mut['Codon_Mut'].append(mut_codon)
                    self.dict_mut['RTT_DNA_frame'].append(rtt_frame)
                    self.dict_mut['RTT_DNA_Strand'].append(strand)
                    self.dict_mut['AminoAcid_WT'].append(aa_wt)
                    self.dict_mut['AminoAcid_Mut'].append(aa_mut)
                    self.dict_mut['Silent_check'].append(silent_check)
                    self.dict_mut['Mut_pos'].append(mut_pos)
                    self.dict_mut['Mut_refpos'].append(mut_refpos)
                    self.dict_mut['PAM_Mut'].append(rtt_dna_mut[4:6])
                    self.dict_mut['RTT_DNA_Mut'].append(rtt_dna_mut)
                    self.dict_mut['Edit_class'].append('RHA_edit')

        self.mutations  = pd.DataFrame(self.dict_mut) 

        df_synonymous = self.mutations.groupby(by='Edit_class').get_group('RHA_edit').reset_index(drop=True)
        df_synonymous = df_synonymous.groupby(by='Silent_check').get_group(True).sort_values(by='Priority').reset_index(drop=True)

        return df_synonymous

    # def End: make_synonyRHA


def mismatch(seq: str, 
             n: int, 
             start: int = 0, 
             end: int = -1, 
             capital: bool = False,
             full: bool = False,
             ):
    
    '''
    seq  : mismatch를 만들고자 하는 sequence 정보 (DNA 기준, 추후 RNA 추가해주면 좋을듯?)
    n    : mismatch를 만드는 수
    start: target에서 mismatch를 도입할 시작점
    end  : target에서 mismatch를 도입할 종료점
    capital: mismatched nucleotide 표기를 대문자로 할 것인지, True이면 대문자로 표시됨
    full: 모든 mismatched position, WT, Alt, original seq 등 자세한 내용을 DataFrame으로 받을지.
    '''
    
    from itertools import combinations, product
    
    '''
    아직 미완성!!!!
    '''
    
    seq = seq.upper()
    target_seq = seq[start:end]
    
    input_len = len(seq)
    list_seq = list(seq)
    dic = {}
    loc = list(combinations(range(input_len), n))
    nucleo_dic = {"A": ["T","G","C"], 
                  "T": ["A","G","C"], 
                  "G": ["A","T","C"], 
                  "C": ["A","T","G"]}
    
    for i in loc:
        b = list_seq.copy()
        for k in range(len(i)):
            b[i[k]] = nucleo_dic[b[i[k]]]
        lst = list(product(*b))
        for i in lst:
            dic [''.join(i)] = input
            
    return dic
    

class pegRNA:
    '''
    Dev-ing...
    ToDo: RT-PBS combination dict -> DataFrame transformation
    PBS / RTT separated information must included.
    '''
    def __init__(self, wt_seq, ed_seq, edit_type, edit_len,
                pbs_min:int=6, pbs_max:int=17, rtt_max=40,
                pe_system='PE2max'):
        
        self.sWTSeq = wt_seq
        self.sEditedSeq = ed_seq
        self.sAltKey = edit_type + str(edit_len)
        self.sAltType = edit_type
        self.nAltLen = edit_len

        self.nAltIndex = 60
        self.pbs_range = [pbs_min, pbs_max]
        self.rtt_max   = rtt_max
        self.pe_system = pe_system

        self.sGuideKey = ''
        self.sChrID = ''
        self.sStrand = ''
        self.nGenomicPos = 0
        self.nEditIndex = 0
        self.nPBSLen = 0
        self.nRTTLen = 0
        self.sPBSSeq = ''
        self.sRTSeq = ''
        self.sPegRNASeq = ''
        self.list_sSeqs = []
        self.type_sub = 0
        self.type_ins = 0
        self.type_del = 0
        self.dict_sSeqs = {}
        self.dict_sCombos = {}
        self.dict_sOutput = {}

        if   self.sAltType.startswith('sub'): self.type_sub = 1
        elif self.sAltType.startswith('del'): self.type_del = 1
        elif self.sAltType.startswith('ins'): self.type_ins = 1

        
        self.get_sAltNotation()
        self.get_all_RT_PBS(nMinPBS=self.pbs_range[0]-1, nMaxPBS=self.pbs_range[1], nMaxRT=self.rtt_max, pe_system=pe_system)
        self.make_rt_pbs_combinations()

        self.df_out = pd.DataFrame(self.dict_sCombos)

    # def End: get_input


    def show_output(self): return self.df_out

    def get_sAltNotation(self):
        if self.sAltType == 'sub':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[self.nAltIndex:self.nAltIndex + self.nAltLen], self.sEditedSeq[self.nAltIndex:self.nAltIndex + self.nAltLen])

        elif self.sAltType == 'del':
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[self.nAltIndex:self.nAltIndex + 1 + self.nAltLen], self.sEditedSeq[self.nAltIndex])

        else:
            self.sAltNotation = '%s>%s' % (
                self.sWTSeq[self.nAltIndex], self.sEditedSeq[self.nAltIndex:self.nAltIndex + self.nAltLen + 1])

    # def END: get_sAltNotation

    def get_all_RT_PBS(self, 
                    nMinPBS = 0,
                    nMaxPBS = 17,
                    nMaxRT = 40,
                    nSetPBSLen = 0,
                    nSetRTLen = 0,
                    pe_system = 'PE2'
                    ):
        """
        nMinPBS: If you set specific number, lower than MinPBS will be not generated. Default=0
        nMaxPBS: If you set specific number, higher than MinPBS will be not generated. Default=17
        nMaxRT = : If you set specific number, higher than MinPBS will be not generated. Default=40
        nSetPBSLen = 0  # Fix PBS Len: Set if >0
        nSetRTLen = 0  # Fix RT  Len: Set if >0
        PAM: 4-nt sequence
        """

        nMaxEditPosWin = nMaxRT + 3  # Distance between PAM and mutation

        dict_sWinSize = {'sub': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 2 - 3, 6], 3: [nMaxRT - 3 - 3, 6]},
                        'ins': {1: [nMaxRT - 2 - 3, 6], 2: [nMaxRT - 3 - 3, 6], 3: [nMaxRT - 4 - 3, 6]},
                        'del': {1: [nMaxRT - 1 - 3, 6], 2: [nMaxRT - 1 - 3, 6], 3: [nMaxRT - 1 - 3, 6]}}

        
        if 'NRCH' in pe_system: # for NRCH-PE PAM
            dict_sRE = {'+': '[ACGT][ACGT]G[ACGT]|[ACGT][CG]A[ACGT]|[ACGT][AG]CC|[ATCG]ATG', 
                        '-': '[ACGT]C[ACGT][ACGT]|[ACGT]T[CG][ACGT]|G[GT]T[ACGT]|ATT[ACGT]|CAT[ACGT]|GGC[ACGT]|GTA[ACGT]'} 
        else:
            dict_sRE = {'+': '[ACGT]GG[ACGT]', '-': '[ACGT]CC[ACGT]'} # for Original-PE PAM

        for sStrand in ['+', '-']:

            sRE = dict_sRE[sStrand]
            for sReIndex in regex.finditer(sRE, self.sWTSeq, overlapped=True):

                if sStrand == '+':
                    nIndexStart = sReIndex.start()
                    nIndexEnd = sReIndex.end() - 1
                    sPAMSeq = self.sWTSeq[nIndexStart:nIndexEnd]
                    sGuideSeq = self.sWTSeq[nIndexStart - 20:nIndexEnd]
                else:
                    nIndexStart = sReIndex.start() + 1
                    nIndexEnd = sReIndex.end()
                    sPAMSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd])
                    sGuideSeq = reverse_complement(self.sWTSeq[nIndexStart:nIndexEnd + 20])

                nAltPosWin = set_alt_position_window(sStrand, self.sAltKey, self.nAltIndex, nIndexStart, nIndexEnd,
                                                    self.nAltLen)

                ## AltPosWin Filter ##
                if nAltPosWin <= 0:             continue
                if nAltPosWin > nMaxEditPosWin: continue

                nPAM_Nick = set_PAM_nicking_pos(sStrand, self.sAltType, self.nAltLen, self.nAltIndex, nIndexStart, nIndexEnd)

                if not check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, self.sAltType, self.nAltLen,
                                        self.nAltIndex): continue

                sPAMKey = '%s,%s,%s,%s,%s,%s,%s' % (
                    self.sAltKey, self.sAltNotation, sStrand, nPAM_Nick, nAltPosWin, sPAMSeq, sGuideSeq)

                dict_sRT, dict_sPBS = self.determine_PBS_RT_seq(sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen,
                                                        nSetRTLen, nPAM_Nick, nAltPosWin, self.sEditedSeq)

                nCnt1, nCnt2 = len(dict_sRT), len(dict_sPBS)
                if nCnt1 == 0: continue
                if nCnt2 == 0: continue
                
                if sPAMKey not in self.dict_sSeqs:
                    self.dict_sSeqs[sPAMKey] = ''
                self.dict_sSeqs[sPAMKey] = [dict_sRT, dict_sPBS]

            # loop END: sReIndex
        # loop END: sStrand


    # def END: get_all_RT_PBS

    def determine_PBS_RT_seq(self, sStrand, nMinPBS, nMaxPBS, nMaxRT, nSetPBSLen, nSetRTLen, nPAM_Nick,
                            nAltPosWin, sForTempSeq):
        dict_sPBS = {}
        dict_sRT = {}

        list_nPBSLen = [nNo + 1 for nNo in range(nMinPBS, nMaxPBS)]
        for nPBSLen in list_nPBSLen:

            ## Set PBS Length ##
            if nSetPBSLen:
                if nPBSLen != nSetPBSLen: continue

            if sStrand == '+':
                nPBSStart = nPAM_Nick - nPBSLen  # 5' -> PamNick
                nPBSEnd = nPAM_Nick
                sPBSSeq = sForTempSeq[nPBSStart:nPBSEnd] # sForTempSeq = self.EditedSeq

            else:
                if self.sAltKey.startswith('sub'):
                    nPBSStart = nPAM_Nick
                elif self.sAltKey.startswith('ins'):
                    nPBSStart = nPAM_Nick + self.nAltLen
                elif self.sAltKey.startswith('del'):
                    nPBSStart = nPAM_Nick - self.nAltLen

                sPBSSeq = reverse_complement(sForTempSeq[nPBSStart:nPBSStart + nPBSLen]) # sForTempSeq = self.EditedSeq

            # if END: sStrand

            sKey = len(sPBSSeq)
            if sKey not in dict_sPBS:
                dict_sPBS[sKey] = ''
            dict_sPBS[sKey] = sPBSSeq
        # loop END: nPBSLen

        if sStrand == '+':
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo + 1 for nNo in range(self.nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            elif self.sAltKey.startswith('ins'):
                list_nRTPos = [nNo + 1 for nNo in range(self.nAltIndex + self.nAltLen, (nPAM_Nick + nMaxRT))] # OK
            else:
                list_nRTPos = [nNo + 1 for nNo in range(self.nAltIndex, (nPAM_Nick + nMaxRT))] ## 수정! ## del2 RHA 3 del1 RHA2
        else:
            if self.sAltKey.startswith('sub'):
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 1 - nMaxRT, self.nAltIndex)] ## 수정! ## sub1 sub 3 RHA 0
            else:
                list_nRTPos = [nNo for nNo in range(nPAM_Nick - 3 - nMaxRT, self.nAltIndex + self.nAltLen - 1)] ## 수정! ## ins2 최소가 2까지 ins3 RHA 최소 3 #del2 RHA 2 del1 RHA1
        for nRTPos in list_nRTPos:

            if sStrand == '+':
                nRTStart = nPAM_Nick  # PamNick -> 3'
                nRTEnd = nRTPos
                sRTSeq = sForTempSeq[nRTStart:nRTEnd]

            else:
                if self.sAltKey.startswith('sub'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick  # PamNick -> 3'
                elif self.sAltKey.startswith('ins'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick + self.nAltLen  # PamNick -> 3'
                elif self.sAltKey.startswith('del'):
                    nRTStart = nRTPos
                    nRTEnd = nPAM_Nick - self.nAltLen  # PamNick -> 3'

                sRTSeq = reverse_complement(sForTempSeq[nRTStart:nRTEnd])

                if not sRTSeq: continue
            # if END: sStrand

            sKey = len(sRTSeq)

            ## Set RT Length ##
            if nSetRTLen:
                if sKey != nSetRTLen: continue

            ## Limit Max RT len ##
            if sKey > nMaxRT: continue

            ## min RT from nick site to mutation ##
            if self.sAltKey.startswith('sub'):
                if sStrand == '+':
                    if sKey < abs(self.nAltIndex - nPAM_Nick): continue
                else:
                    if sKey < abs(self.nAltIndex - nPAM_Nick + self.nAltLen - 1): continue ### 
            else:
                if sStrand == '-':
                    if sKey < abs(self.nAltIndex - nPAM_Nick + self.nAltLen - 1): continue

            if self.sAltKey.startswith('ins'):
                if sKey < nAltPosWin + 1: continue

            if sKey not in dict_sRT:
                dict_sRT[sKey] = ''
            dict_sRT[sKey] = sRTSeq
        # loop END: nRTPos

        return [dict_sRT, dict_sPBS]


    # def END: determine_PBS_RT_seq

    def make_rt_pbs_combinations(self):
        for sPAMKey in self.dict_sSeqs:

            dict_sRT, dict_sPBS = self.dict_sSeqs[sPAMKey]

            list_sRT = [dict_sRT[sKey] for sKey in dict_sRT]
            list_sPBS = [dict_sPBS[sKey] for sKey in dict_sPBS]

            if sPAMKey not in self.dict_sCombos:
                self.dict_sCombos[sPAMKey] = ''
            self.dict_sCombos[sPAMKey] = {'%s,%s' % (sRT, sPBS): {} for sRT in list_sRT for sPBS in list_sPBS}
        # loop END: sPAMKey
    # def END: make_rt_pbs_combinations


def reverse_complement(sSeq):
    dict_sBases = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N': 'N', 'U': 'U', 'n': '',
                   '.': '.', '*': '*', 'a': 't', 'c': 'g', 'g': 'c', 't': 'a'}
    list_sSeq = list(sSeq)  # Turns the sequence in to a gigantic list
    list_sSeq = [dict_sBases[sBase] for sBase in list_sSeq]
    return ''.join(list_sSeq)[::-1]

# def END: reverse_complement

def set_alt_position_window(sStrand, sAltKey, nAltIndex, nIndexStart, nIndexEnd, nAltLen):
    if sStrand == '+':

        if sAltKey.startswith('sub'):
            return (nAltIndex + 1) - (nIndexStart - 3)
        else:
            return (nAltIndex + 1) - (nIndexStart - 3)

    else:
        if sAltKey.startswith('sub'):
            return nIndexEnd - nAltIndex + 3 - (nAltLen - 1)

        elif sAltKey.startswith('del'):
            return nIndexEnd - nAltIndex + 3 - nAltLen

        else:
            return nIndexEnd - nAltIndex + 3 + nAltLen
        # if END:
    # if END:

# def END: set_alt_position_window


def set_PAM_nicking_pos(sStrand, sAltType, nAltLen, nAltIndex, nIndexStart, nIndexEnd):
    if sStrand == '-':
        nPAM_Nick = nIndexEnd + 3
    else:
        nPAM_Nick = nIndexStart - 3

    return nPAM_Nick

# def END: set_PAM_Nicking_Pos


def check_PAM_window(dict_sWinSize, sStrand, nIndexStart, nIndexEnd, sAltType, nAltLen, nAltIndex):
    nUp, nDown = dict_sWinSize[sAltType][nAltLen]

    if sStrand == '+':
        nPAMCheck_min = nAltIndex - nUp + 1
        nPAMCheck_max = nAltIndex + nDown + 1
    else:
        nPAMCheck_min = nAltIndex - nDown + 1
        nPAMCheck_max = nAltIndex + nUp + 1
    # if END:

    if nIndexStart < nPAMCheck_min or nIndexEnd > nPAMCheck_max:
        return 0
    else:
        return 1

# def END: check_PAM_window


import os, requests
import pandas as pd
import genet.database as db
import genet.utils as U
from ftplib import FTP

class NCBI(db.config.NCBIconfig):
    def __init__(self, ):
        """NCBI FTP 서버에 접속해서 원하는 metadata를 다운로드 받고,
        metadata를 기반으로 원하는 genome data parsing을 할 수 있는 module
        """        

        super().__init__()

        if self.isexist:
            self.meta = pd.read_parquet(f'{self.local_path}/{self.local_file}')
        else:
            print('[Info] NCBI reference genome assembly metadata is not found. This message appears only once when starting the NCBI database for the first time.')
            self.download(download_path=self.local_path)
            self.meta = pd.read_parquet(f'{self.local_path}/{self.local_file}')

    
    # def End: __init__
        
    def download(self, download_path:str, convert=True):
        '''Download files from FTP server to local path.
        If download_path pointed, metadata file will be downloaded at pointed path.
        
        ToDo
        GetGenome이 이 NCBI를 상속 받는데, 똑같은 이름의 method (download)가 있음.
        헷갈릴 수 있으니, 이름을 서로 다르게 하거나 NCBI의 download method를 좀 바꿔야 할 것 같음.
        '''

        print('[Info] Downloading NCBI assembly summary of reference sequence')

        U.request_file(
            server      = self.ftp_server,
            remote_path = self.remote_path,
            local_path  = download_path, 
            target_file = self.target_file, 
            )
        
        # File will be converted to parquet format automatically
        if convert==True: self._convert_to_parquet()
        
        print(f'[Info] Complete')

    # def End: download
            
    def _convert_to_parquet(self):
        '''txt 파일은 너무 무겁고 data loading이 느린 것 같아서, 
        pandas에서 빠르게 읽고 용량도 가벼운 parquet (파케이) 형식으로 변환.
        처음 한번만 변환하면, 나중에 언제든 불러올 때 매우 빠르다. 
        '''        

        original_file  = f'{self.local_path}/assembly_summary_refseq.txt'
        converted_file = f'{self.local_path}/assembly_summary_refseq.parquet'

        print('[Info] Converting format: .txt to .parquet')
        meta = pd.read_csv(original_file, sep='\t', header=1, low_memory=False)
        meta.to_parquet(converted_file, index=False)

        # remove original text file
        os.remove(original_file)

        print(f'[Info] Removed summary text file: {original_file}')
        print(f'[Info] Converted metadata file: {converted_file}')

    # def End: _convert_to_parquet

    def update(self):
        '''Update files from FTP server to local path.
        '''

        print('[Info] Updating NCBI assembly summary of reference sequence')

        U.request_file(
            server      = self.ftp_server,
            remote_path = self.remote_path,
            local_path  = self.local_path, 
            target_file = self.target_file, 
            )
        
        # File will be converted to parquet format automatically
        self._convert_to_parquet()
        
        print(f'[Info] Complete')

    # def End: download
# class End: NCBI


class GetGenome(NCBI):
    def __init__(self, id:str, category:str='organism'):
        '''Metadata의 category에 속하는 것들 중 선택된 것에 맞는 id를 찾아서
        정보를 불러오고, 다운로드 해주는 함수.
        '''
        super().__init__()
        
        self.category_available = ['accession', 'organism']

        if category not in self.category_available: 
            raise ValueError('''[Error] Not valid category. Please check your category input.
                    Available categories: "accession" or "organism"''')
        
        if   category == 'organism' : category = 'organism_name'
        elif category == 'accession': category = '#assembly_accession'
        
        # 카테고리로 지정된 column을 index로 지정한 dataframe
        _columns  = self.meta.columns
        self.meta = self.meta.set_index(category)

        try   : self.data = self.meta.loc[[id]].reset_index()[_columns]
        except: raise ValueError('''[Error] Not valid ID, Please check your id or category ("organism" or "accession") input.''')

        if category == 'organism_name':
            self.info = self._search_refseq()
        else:
            self.info = self.data.loc[0]
        
        del self.meta
    
    # def End: __init__

    def _search_refseq(self, ) -> pd.Series():
        """만약 category가 organism_name이라면, 여러개의 GCF들이 나올 수 있으니,
        그 중에서 어떤 것이 ref_seq인지 선정하기 위한 method.
        만약 적절한 ref_seq이 없다면, 다른 category를 이용하는 것을 권장하는 error 발생.

        Raises:
            ValueError: 해당 organism에서 ref_seq으로 지정된 것이 하나도 없는 경우.
            ValueError: 해당 organism에서 ref_seq으로 지정된 것이 여러 개인 경우.

        Returns:
            _type_: _description_
        """        
             
        list_access  = list(self.data['#assembly_accession'])
        ref_category = list(self.data.refseq_category)
        index_refseq = [index for index, value in enumerate(ref_category) if value != 'na']

        if   len(index_refseq) == 1:
            return self.data.iloc[index_refseq[0]]
        
        elif len(index_refseq) == 0:
            self.data.to_csv('test.csv')
            raise ValueError(f'''[Info] There are no defined reference genome. 
                    Please use "#assembly_accession" as category.
                    You should select specific genome depending on your research purpose.
                    Available accessions: {list_access}''')
        
        elif len(index_refseq) > 1:
            raise ValueError(f'''[Info] There are more than one defined reference genome. 
                    Please use "#assembly_accession" as category.
                    Available accessions: {list_access}''')
            
    # def End: _search_refseq
    
    def __call__(self,):
        return self.info
    
    def contents(self) -> list:
        """NCBI FTP 서버에 특정 spacies 폴더 안에 들어있는 파일들의 list를 전부 뽑아주는 함수. 
        list 형태의 결과로 return 한다.

        Returns:
            list: List of files and directories in NCBI FTP server at pointed path.
        """        
        
        ftp_path = self.info.ftp_path
        paths = ftp_path.split('//')[1]
        
        server, remote_path = paths.split('/', 1)
        
        # FTP Connect
        with FTP(self.ftp_server) as ftp:
            ftp.login()
            
            ftp.cwd(remote_path)
            list_files = ftp.nlst()
            
        return list_files
    
    # def End: contents

    def download(self, target_file:str, path:str='./', silence=False):
                
        """_summary_

        Args:
            target_file (str): File name for download from NCBI server.
            path (str, optional): Local path for save downloaded file. Defaults to './' (current working directory).
        """                
        ftp_path = self.info.ftp_path
        paths = ftp_path.split('//')[1]
        
        server, remote_path = paths.split('/', 1)

        try:
            U.request_file(
                server      = server,
                remote_path = remote_path,
                local_path  = path, 
                target_file = target_file, 
                silence     = silence,
            )
        
        except:
            print(f'[Error] Fail to download file. Available file: {self.contents()}')

    # def End: download


class GenBankParser:
    def __init__(self, gb_file:str, feature_file:str):
        """GenBank file과 feature file을 같이 넣어줘서 feature 파일 내에 있는 정보를 기반으로 parsing 하는 함수.
        특히 transcript는 "tag=MANE Select" 가 들어있는 경우가 representative mRNA라는 정보를 가져오는 것이 중요하다.
        이에 대해서는 좀 더 고민해보고 class 만들어보자.

        Args:
            gb_file (str): .gbff 파일
            feature_file (str): .gff 파일
        """        

        print('[Info] Parsing GenBank file.')

    # def End: __init__






    


class FrameConverter:
    def __init__(self):
        
        self.available_format = [
            '.gff', 'gff.gz',
            '.gtf', 'gtf.gz',
        ]

    # def End: __init__


    def convert(self, file_path) -> pd.DataFrame:
        """Database에서 받은 각종 파일들을 dataframe으로 바꿔주는 method.

        Args:
            file_path (_type_): 변환하고 싶은 파일의 경로.

        Raises:
            ValueError: Format을 지원하지 않는 파일이 들어왔을 때 Error.

        Returns:
            pd.DataFrame: Data file을 DataFrame으로 변환한 것. 
        """        

        for fmt in self.available_format:
            if file_path.endswith(fmt): break

            raise ValueError(f'Not available format. Available: {self.available_format}')
        
        if fmt in ['.gff', 'gff.gz', '.gtf', 'gtf.gz']:
            df = self._gff2df(file_path)

        
        return df
        
    # def End: convert
        



    def _gff2df(self, file_path:str) -> pd.DataFrame:
        """GFF3 파일을 pd.DataFrame에 담아주는 함수.

        Args:
            file_path (str): gff 파일의 경로. 또는 gzipped file도 가능하다 (.gff.gz)

        Returns:
            pd.DataFrame: gff (or gtf) 파일의 정보가 담긴 dataframe
        """        
        
        col_names = ['seqid', 'source', 'type', 'start', 'end', 'score', 'strand', 'phase', 'attributes']
        df_gff = pd.read_csv(file_path, sep='\t', names=col_names, comment='#')
        
        return df_gff
    
    # def End: _gff2df







import genet
import sys
import pandas as pd
from glob import glob
from tqdm import tqdm


class CountReadFromCRISPResso:
    def __init__(self, exon_num, ref_path:str, wt_seq:str):
        
        # __init__()
        self.exon_num = str(exon_num)
        self.df_ref   = pd.read_csv(ref_path)
        
        if  self.exon_num in ['4', '5', '6', '7', '8', '9']:
            self.label_synpe  = 'SynPE ref\nInput'    # Intended edit + Synony edit (perfect edit)
            self.label_intend = 'Intended Only Input' # Intended edit only
            self.label_synony = 'Synony Only Input'   # Synonymous edit only
            self.df_ref_ex = self.df_ref[self.df_ref['Exon']==self.exon_num]
        
        else:
            print('Input error: Please check your exon number.')
            sys.exit()
        
        self.df_ref_ex = self.df_ref_ex.reset_index(drop=True)
        self.wt_refseq = wt_seq


    def run(self, cs_file:str) -> pd.DataFrame:
            
        df     = pd.read_csv(cs_file, sep = '\t')
        f_name = cs_file.split('\\')[-1].split('.txt')[0]

        # Step1: read CRISPResso aligned & reference file
        
        dict_out  = {'No_matched': 0,
                    self.wt_refseq: 0,}

        for i in self.df_ref_ex.index:
            ref = self.df_ref_ex.iloc[i]
            
            seq_synpe  = ref[self.label_synpe]    # Intended edit + Synony edit (perfect edit)
            
            dict_out[seq_synpe]  = 0

        print(f'\nStart - {f_name}')
        print('Length of Dict:', len(dict_out))

        # Step2: read count
        for i in tqdm(df.index, desc=f'Read counting: {f_name}'):
            data = df.iloc[i]
            seq  = data['Aligned_Sequence']
            cnt  = int(data['#Reads'])
            
            try   : dict_out[seq] += cnt
            except: dict_out['No_matched'] += cnt
            
            
        # Step3: make output file
        list_synpe  = []

        for i in self.df_ref_ex.index:
            ref = self.df_ref_ex.iloc[i]
            
            seq_synpe  = ref[self.label_synpe]    # Intended edit + Synony edit (perfect edit)
            
            list_synpe.append(dict_out[seq_synpe])
            
        df_out = self.df_ref_ex.copy()
        df_out['SynPE_count'] = list_synpe

        # Step4: drop duplicates and save output file
        df_out_nodupl = df_out.copy()
        df_out_nodupl = df_out_nodupl.drop_duplicates([self.label_synpe])

        print('No_matched:', dict_out['No_matched'])
        print('WT_refseq :', dict_out[self.wt_refseq])

        return df_out_nodupl



'''Utilities.py - Tools for writing reproducible scripts
=========================================================

The :mod:`Utilities` modules contains utility functions for argument
parsing, logging and record keeping within scripts.

This module is imported by most UMI-tools tools. It provides convenient
and consistent methods for

   * `Record keeping`_
   * `Argument parsing`_
   * `Input/Output redirection`_
   * `Logging`_
   * `Running external commands`_
   * `Benchmarking`_

The basic usage of this module within a script is::

    """script_name.py - my script

    Mode Documentation
    """
    import sys
    import optparse
    import Utilities as U

    def main(argv=None):
        """script main.

        parses command line options in sys.argv, unless *argv* is given.
        """

        if not argv: argv = sys.argv

        # setup command line parser
        parser = U.OptionParser(version="%prog version: $Id$",
                                usage=globals()["__doc__"] )

        parser.add_option("-t", "--test", dest="test", type="string",
                          help="supply help")

        # add common options (-h/--help, ...) and parse
        # command line
        (options, args) = U.Start(parser)

        # do something
        # ...
        U.info("an information message")
        U.warn("a warning message)

        ## write footer and output benchmark information.
        U.Stop()

    if __name__ == "__main__":
        sys.exit(main(sys.argv))

Record keeping
--------------

The central functions in this module are the :py:func:`Start` and
:py:func:`Stop` methods which are called before or after any work is
done within a script.

The :py:func:`Start` is called with an U.OptionParser object.
:py:func:`Start` will add additional command line arguments, such as
``--help`` for command line help or ``--verbose`` to control the
:term:`loglevel`.  It can also add optional arguments for scripts
needing database access, writing to multiple output files, etc.

:py:func:`Start` will write record keeping information to a
logfile. Typically, logging information is output on stdout, prefixed
by a `#`, but it can be re-directed to a separate file. Below is a
typical output::

    # output generated by /ifs/devel/andreas/cgat/beds2beds.py --force-output --exclusive-overlap --method=unmerged-combinations --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz Rela-030m-R1.bed.gz
    # job started at Thu Mar 29 13:06:33 2012 on cgat150.anat.ox.ac.uk -- e1c16e80-03a1-4023-9417-f3e44e33bdcd
    # pid: 16649, system: Linux 2.6.32-220.7.1.el6.x86_64 #1 SMP Fri Feb 10 15:22:22 EST 2012 x86_64
    # exclusive                               : True
    # filename_update                         : None
    # ignore_strand                           : False
    # loglevel                                : 1
    # method                                  : unmerged-combinations
    # output_filename_pattern                 : 030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz
    # output_force                            : True
    # pattern_id                              : (.*).bed.gz
    # stderr                                  : <open file \'<stderr>\', mode \'w\' at 0x2ba70e0c2270>
    # stdin                                   : <open file \'<stdin>\', mode \'r\' at 0x2ba70e0c2150>
    # stdlog                                  : <open file \'030m.intersection.tsv.log\', mode \'a\' at 0x1f1a810>
    # stdout                                  : <open file \'<stdout>\', mode \'w\' at 0x2ba70e0c21e0>
    # timeit_file                             : None
    # timeit_header                           : None
    # timeit_name                             : all
    # tracks                                  : None

The header contains information about:

    * the script name (``beds2beds.py``)

    * the command line options (``--force-output --exclusive-overlap
      --method=unmerged-combinations
      --output-filename-pattern=030m.intersection.tsv.dir/030m.intersection.tsv-%s.bed.gz
      --log=030m.intersection.tsv.log Irf5-030m-R1.bed.gz
      Rela-030m-R1.bed.gz``)

    * the time when the job was started (``Thu Mar 29 13:06:33 2012``)
    * the location it was executed (``cgat150.anat.ox.ac.uk``)
    * a unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)
    * the pid of the job (``16649``)

    * the system specification (``Linux 2.6.32-220.7.1.el6.x86_64 #1
      SMP Fri Feb 10 15:22:22 EST 2012 x86_64``)

It is followed by a list of all options that have been set in the script.

Once completed, a script will call the :py:func:`Stop` function to
signify the end of the experiment.

:py:func:`Stop` will output to the log file that the script has
concluded successfully. Below is typical output::

    # job finished in 11 seconds at Thu Mar 29 13:06:44 2012 -- 11.36 0.45 0.00 0.01 -- e1c16e80-03a1-4023-9417-f3e44e33bdcd

The footer contains information about:

   * the job has finished (``job finished``)
   * the time it took to execute (``11 seconds``)
   * when it completed (``Thu Mar 29 13:06:44 2012``)
   * some benchmarking information (``11.36  0.45  0.00  0.01``)
     which is ``user time``, ``system time``,
     ``child user time``, ``child system time``.
   * the unique job id (``e1c16e80-03a1-4023-9417-f3e44e33bdcd``)

The unique job id can be used to easily retrieve matching information
from a concatenation of log files.

Argument parsing
----------------

The module provides :class:`OptionParser` to facilitate option
parsing.  :class:`OptionParser` is derived from the
:py:class:`optparse.OptionParser` class, but has improvements to
provide better formatted output on the command line. It also allows to
provide a comma-separated list to options that accept multiple
arguments. Thus, ``--method=sort --method=crop`` and
``--method=sort,crop`` are equivalent.

Input/Output redirection
------------------------

:func:`Start` adds the options ``--stdin``, ``--stderr` and
``--stdout`` which allow using files as input/output streams.

To make this work, scripts should not read from sys.stdin or write to
sys.stdout directly, but instead use ``options.stdin`` and
``options.stdout``. For example to simply read all lines from stdin
and write to stdout, use::

   (options, args) = U.Start(parser)

   input_data = options.stdin.readlines()
   options.stdout.write("".join(input_data))

The script can then be used in many different contexts::

   cat in.data | python script.py > out.data
   python script.py --stdin=in.data > out.data
   python script.py --stdin=in.data --stdout=out.data

The method handles gzip compressed files transparently. The following
are equivalent::

   zcat in.data.gz | python script.py | gzip > out.data.gz
   python script.py --stdin=in.data.gz --stdout=out.data.gz

For scripts producing multiple output files, use the argument
``add_output_options=True`` to :func:`Start`. This provides the option
``--output-filename-pattern`` on the command line. The user can then
supply a pattern for output files. Any ``%s`` appearing in the pattern
will be substituted by a ``section``. Inside the script, When opening
an output file, use the method :func:`openOutputFile` to provide a
file object::

   output_histogram = U.openOutputFile(section="histogram")
   output_stats = U.openOutputFile(section="stats")

If the user calls the script with::

   python script.py --output-filename-pattern=sample1_%s.tsv.gz

the script will create the files ``sample1_histogram.tsv.gz`` and
``sample1_stats.tsv.gz``.

This method will also add the option ``--force-output`` to permit
overwriting existing files.

Logging
-------

:py:mod:`Utilities` provides the well known logging methods from
the :py:mod:`logging` module such as :py:func:`info`,
:py:func:`warn`, etc. These are provided so that no additional import
of the :py:mod:`logging` module is required, but either functions
can be used.

Running external commands
-------------------------

The :func:`run` method is a shortcut :py:func:`subprocess.call` and
similar methods with some additional sanity checking.

Benchmarking
------------

The :func:`Start` method records basic benchmarking information when a
script starts and :func:`Stop` outputs it as part of its final log
message::

    # job finished in 11 seconds at Thu Mar 29 13:06:44 2012 -- 11.36 0.45 0.00 0.01 -- e1c16e80-03a1-4023-9417-f3e44e33bdcd

See `Record keeping`_ for an explanations of the fields.

To facilitate collecting benchmark information from running multiple
scripts, these data can be tagged and saved in a separate file. See the
command line options ``--timeit``, ``--timeit-name``, ``--timeit-header``
in :func:`Start`.

The module contains some decorator functions for benchmarking
(:func:`benchmark`) and caching function (:func:`cachedfunction`) or
class method (:func:`cachedmethod`) calls.

Complete reference
------------------

'''

########################################################################################
# The code for Utilities.py has been taken with permission from CGAT.Experiment.py
# https://github.com/CGATOxford/cgat/blob/master/CGAT/Experiment.py
# and CGATPipelines.CGATPipelines.Pipeline.Files.py
# https://github.com/CGATOxford/CGATPipelines/blob/master/CGATPipelines/Pipeline/Files.py
#########################################################################################


import re
import sys
import time
import inspect
import copy
import os
import logging
import collections
import gzip
import optparse
import textwrap
import random
import uuid
import tempfile
import regex
import itertools

from builtins import bytes, chr


class DefaultOptions:
    stdlog = sys.stdout
    stdout = sys.stdout
    stderr = sys.stderr
    stdin = sys.stdin
    loglevel = 2
    timeit_file = None
    compresslevel = 6

global_starting_time = time.time()
global_options = DefaultOptions()
global_args = None
global_id = uuid.uuid4()
global_benchmark = collections.defaultdict(int)

##########################################################################
# The code for BetterFormatter has been taken from
# http://code.google.com/p/yjl/source/browse/Python/snippet/BetterFormatter.py
__copyright__ = """
Copyright (c) 2001-2006 Gregory P. Ward.  All rights reserved.
Copyright (c) 2002-2006 Python Software Foundation.  All rights reserved.
Copyright (c) 2011 Yu-Jie Lin.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

  * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

  * Neither the name of the author nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


class BetterFormatter(optparse.IndentedHelpFormatter):
    """A formatter for :class:`OptionParser` outputting indented
    help text.
    """

    def __init__(self, *args, **kwargs):

        optparse.IndentedHelpFormatter.__init__(self, *args, **kwargs)
        self.wrapper = textwrap.TextWrapper(width=self.width)

    def _formatter(self, text):

        return '\n'.join(['\n'.join(p) for p in
                          map(self.wrapper.wrap,
                              self.parser.expand_prog_name(text).split('\n'))])

    def format_description(self, description):

        if description:
            return self._formatter(description) + '\n'
        else:
            return ''

    def format_epilog(self, epilog):

        if epilog:
            return '\n' + self._formatter(epilog) + '\n'
        else:
            return ''

    def format_usage(self, usage):

        msg = '''

For full UMI-tools documentation, see https://umi-tools.readthedocs.io/en/latest/\n'''

        return optparse._(usage) + msg

    def format_option(self, option):

        # Ripped and modified from Python 2.6's optparse's HelpFormatter
        result = []
        opts = self.option_strings[option]
        opt_width = self.help_position - self.current_indent - 2
        if len(opts) > opt_width:
            opts = "%*s%s\n" % (self.current_indent, "", opts)
            indent_first = self.help_position
        else:                       # start help on same line as opts
            opts = "%*s%-*s  " % (self.current_indent, "", opt_width, opts)
            indent_first = 0
        result.append(opts)

        if option.help:
            help_text = self.expand_default(option)
            # Added expand program name
            help_text = self.parser.expand_prog_name(help_text)
            # Modified the generation of help_line
            help_lines = []
            wrapper = textwrap.TextWrapper(width=self.help_width)
            for p in map(wrapper.wrap, help_text.split('\n')):
                if p:
                    help_lines.extend(p)
                else:
                    help_lines.append('')
            # End of modification
            result.append("%*s%s\n" % (indent_first, "", help_lines[0]))
            result.extend(["%*s%s\n" % (self.help_position, "", line)
                           for line in help_lines[1:]])
        elif opts[-1] != "\n":
            result.append("\n")

        return "".join(result)


# End of BetterFormatter()
#################################################################
#################################################################
#################################################################
class AppendCommaOption(optparse.Option):

    '''Option with additional parsing capabilities.

    * "," in arguments to options that have the action 'append'
      are treated as a list of options. This is what galaxy does,
      but generally convenient.

    * Option values of "None" and "" are treated as default values.
    '''
#    def check_value( self, opt, value ):
# do not check type for ',' separated lists
#        if "," in value:
#            return value
#        else:
#            return optparse.Option.check_value( self, opt, value )
#
#    def take_action(self, action, dest, opt, value, values, parser):
#        if action == "append" and "," in value:
#            lvalue = value.split(",")
#            values.ensure_value(dest, []).extend(lvalue)
#        else:
#            optparse.Option.take_action(
#                self, action, dest, opt, value, values, parser)
#

    def convert_value(self, opt, value):
        if value is not None:
            if self.nargs == 1:
                if self.action == "append":
                    if "," in value:
                        return [self.check_value(opt, v) for v in
                                value.split(",") if v != ""]
                    else:
                        if value != "":
                            return self.check_value(opt, value)
                        else:
                            return value
                else:
                    return self.check_value(opt, value)
            else:
                return tuple([self.check_value(opt, v) for v in value])

    # why is it necessary to pass action and dest to this function when
    # they could be accessed as self.action and self.dest?
    def take_action(self, action, dest, opt, value, values, parser):

        if action == "append" and type(value) == list:
            values.ensure_value(dest, []).extend(value)
        else:
            optparse.Option.take_action(
                self, action, dest, opt, value, values, parser)


class OptionParser(optparse.OptionParser):

    '''UMI-tools derivative of OptionParser.
    '''

    def __init__(self, *args, **kwargs):
        # if "--short" is a command line option
        # remove usage from kwargs
        if "--no-usage" in sys.argv:
            kwargs["usage"] = None

        optparse.OptionParser.__init__(self, *args,
                                       option_class=AppendCommaOption,
                                       formatter=BetterFormatter(),
                                       add_help_option=False,
                                       **kwargs)

        # set new option parser
        # parser.formatter = BetterFormatter()
        # parser.formatter.set_parser(parser)
        if "--no-usage" in sys.argv:
            self.add_option("--no-usage", dest="help_no_usage",
                            action="store_true",
                            help="output help without usage information")


class OptionGroup(optparse.OptionGroup):
    pass


def callbackShortHelp(option, opt, value, parser):
    '''output short help (only command line options).'''
    # clear usage and description
    parser.set_description(None)
    # parser.set_usage(None)
    # output help
    parser.print_help()
    # exit
    parser.exit()


def openFile(filename, mode="r", create_dir=False):
    '''open file in *filename* with mode *mode*.

    If *create* is set, the directory containing filename
    will be created if it does not exist.

    gzip - compressed files are recognized by the
    suffix ``.gz`` and opened transparently.

    Note that there are differences in the file
    like objects returned, for example in the
    ability to seek.

    returns a file or file-like object.
    '''

    _, ext = os.path.splitext(filename)

    if create_dir:
        dirname = os.path.dirname(filename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)

    if ext.lower() in (".gz", ".z"):
        if sys.version_info.major >= 3:
            if mode == "r":
                return gzip.open(filename, 'rt', encoding="ascii")
            elif mode == "w":
                return gzip.open(filename, 'wt',
                                 compresslevel=global_options.compresslevel,
                                 encoding="ascii")
            else:
                raise NotImplementedError(
                    "mode '{}' not implemented".format(mode))
        else:
            return gzip.open(filename, mode,
                             compresslevel=global_options.compresslevel)
    else:
        return open(filename, mode)




def getParams(options=None):
    """return a string containing script parameters.

    Parameters are all variables that start with ``param_``.
    """
    result = []
    if options:
        members = options.__dict__
        for k, v in sorted(members.items()):
            result.append("# %-40s: %s" % (k, str(v)))
    else:
        vars = inspect.currentframe().f_back.f_locals
        for var in filter(lambda x: re.match("param_", x), vars.keys()):
            result.append("# %-40s: %s" %
                          (var, str(vars[var])))

    if result:
        return "\n".join(result)
    else:
        return "# no parameters."


def getFooter():
    """return a header string with command line options and
    timestamp.
    """
    return "# job finished in %i seconds at %s -- %s -- %s" %\
           (time.time() - global_starting_time,
            time.asctime(time.localtime(time.time())),
            " ".join(map(lambda x: "%5.2f" % x, os.times()[:4])),
            global_id)


class MultiLineFormatter(logging.Formatter):

    '''logfile formatter: add identation for multi-line entries.'''

    def format(self, record):
        s = logging.Formatter.format(self, record)
        if s.startswith("#"):
            prefix = "#"
        else:
            prefix = ""
        if record.message:
            header, footer = s.split(record.message)
            s = prefix + s.replace('\n', '\n%s' % prefix + ' ' * len(header))
        return s


def Start(parser=None,
          argv=sys.argv,
          quiet=False,
          add_pipe_options=True,
          add_extract_options=False,
          add_group_dedup_options=True,
          add_sam_options=True,
          add_umi_grouping_options=True,
          return_parser=False):
    """set up an experiment.

    The :py:func:`Start` method will set up a file logger and add some
    default and some optional options to the command line parser.  It
    will then parse the command line and set up input/output
    redirection and start a timer for benchmarking purposes.

    The default options added by this method are:

    ``-v/--verbose``
        the :term:`loglevel`

    ``timeit``
        turn on benchmarking information and save to file

    ``timeit-name``
         name to use for timing information,

    ``timeit-header``
         output header for timing information.

    ``seed``
         the random seed. If given, the python random
         number generator will be initialized with this
         seed.

    Optional options added are:

    Arguments
    ---------

    param parser : :py:class:`U.OptionParser`
       instance with command line options.

    argv : list
        command line options to parse. Defaults to
        :py:data:`sys.argv`

    quiet : bool
        set :term:`loglevel` to 0 - no logging

    return_parser : bool
        return the parser object, no parsing. Useful for inspecting
        the command line options of a script without running it.

    add_pipe_options : bool
        add common options for redirecting input/output

    add_extract_options : bool
        add options for extracting barcodes

    add_sam_options : bool
        add options for SAM/BAM input

    add_umi_grouping_options : bool
        add options for barcode grouping

    add_group_dedup_options : bool
        add options for UMI grouping and deduping

    Returns
    -------
    tuple
       (:py:class:`U.OptionParser` object, list of positional
       arguments)

    """

    if not parser:
        parser = OptionParser(
            version="%prog version: $Id$")

    global global_options, global_args, global_starting_time

    # save default values given by user
    user_defaults = copy.copy(parser.defaults)

    global_starting_time = time.time()

    if add_extract_options:

        group = OptionGroup(parser, "fastq barcode extraction options")

        group.add_option("--extract-method",
                         dest="extract_method", type="choice",
                         choices=["string", "regex"],
                         help=("How to extract the umi +/- cell barcodes, "
                               "Choose from 'string' or 'regex'"))
        group.add_option("-p", "--bc-pattern", dest="pattern", type="string",
                         help="Barcode pattern")
        group.add_option("--bc-pattern2", dest="pattern2", type="string",
                         help="Barcode pattern for paired reads")
        group.add_option("--3prime", dest="prime3", action="store_true",
                         help="barcode is on 3' end of read.")
        group.add_option("--read2-in", dest="read2_in", type="string",
                         help="file name for read pairs")
        group.add_option("--filtered-out",
                         dest="filtered_out", type="string", default=None,
                         help=("Write out reads not matching regex pattern"
                               " to this file"))
        group.add_option("--filtered-out2",
                         dest="filtered_out2", type="string", default=None,
                         help=("Write out paired reads not matching regex"
                               " pattern to this file"))
        group.add_option("--ignore-read-pair-suffixes",
                         dest="ignore_suffix", action="store_true",
                         help="Ignore '\\1' and '\\2' read name suffixes")
        parser.add_option_group(group)

    if add_sam_options:
        group = OptionGroup(parser, "Barcode extraction options")

        group.add_option("--extract-umi-method", dest="get_umi_method", type="choice",
                         choices=("read_id", "tag", "umis"), default="read_id",
                         help="how is the read UMI +/ cell barcode encoded? "
                         "[default=%default]")

        group.add_option("--umi-separator", dest="umi_sep",
                         type="string", help="separator between read id and UMI",
                         default="_")

        group.add_option("--umi-tag", dest="umi_tag",
                         type="string", help="tag containing umi",
                         default='RX')

        group.add_option("--umi-tag-split", dest="umi_tag_split",
                         type="string",
                         help="split UMI in tag and take the first element",
                         default=None)

        group.add_option("--umi-tag-delimiter", dest="umi_tag_delim",
                         type="string",
                         help="concatenate UMI in tag separated by delimiter",
                         default=None)

        group.add_option("--cell-tag", dest="cell_tag",
                         type="string", help="tag containing cell barcode",
                         default=None)

        group.add_option("--cell-tag-split", dest="cell_tag_split",
                         type="string",
                         help=("split cell barcode in tag and take the first"
                               "element for e.g 10X GEM tags"),
                         default='-')

        group.add_option("--cell-tag-delimiter", dest="cell_tag_delim",
                         type="string",
                         help="concatenate cell barcode in tag separated by delimiter",
                         default=None)

        group.add_option("--filter-umi",
                         dest="filter_umi",
                         action="store_true",
                         #help="Filter the UMIs"
                         help=optparse.SUPPRESS_HELP)

        group.add_option("--umi-whitelist", dest="umi_whitelist",
                         type="string", default=None,
                         #help="A whitelist of accepted UMIs"
                         #"[default=%default]"
                         help=optparse.SUPPRESS_HELP)

        group.add_option("--umi-whitelist-paired", dest="umi_whitelist_paired",
                         type="string", default=None,
                         #help="A whitelist of accepted UMIs for "
                         #"read2[default=%default]"
                         help=optparse.SUPPRESS_HELP)

        parser.add_option_group(group)

    if add_umi_grouping_options:
        group = OptionGroup(parser, "UMI grouping options")

        group.add_option("--method", dest="method", type="choice",
                         choices=("adjacency", "directional",
                                  "percentile", "unique", "cluster"),
                         default="directional",
                         help="method to use for umi grouping [default=%default]")

        group.add_option("--edit-distance-threshold", dest="threshold",
                         type="int",
                         default=1,
                         help="Edit distance theshold at which to join two UMIs "
                         "when grouping UMIs. [default=%default]")

        group.add_option("--spliced-is-unique", dest="spliced",
                         action="store_true",
                         help="Treat a spliced read as different to an unspliced"
                         " one [default=%default]",
                         default=False)

        group.add_option("--soft-clip-threshold", dest="soft_clip_threshold",
                         type="float",
                         help="number of bases clipped from 5' end before "
                         "read is counted as spliced [default=%default]",
                         default=4)

        group.add_option("--read-length", dest="read_length",
                         action="store_true", default=False,
                         help="use read length in addition to position and UMI "
                         "to identify possible duplicates [default=%default]")

        parser.add_option_group(group)

    if add_sam_options:
        group = OptionGroup(parser, "single-cell RNA-Seq options")

        group.add_option("--per-gene", dest="per_gene", action="store_true",
                         default=False,
                         help="Group/Dedup/Count per gene. Must combine with "
                         "either --gene-tag or --per-contig")

        group.add_option("--gene-tag", dest="gene_tag",
                         type="string",
                         help="Gene is defined by this bam tag [default=%default]",
                         default=None)

        group.add_option("--assigned-status-tag", dest="assigned_tag",
                         type="string",
                         help="Bam tag describing whether read is assigned to a gene "
                         "By defualt, this is set as the same tag as --gene-tag",
                         default=None)

        group.add_option("--skip-tags-regex", dest="skip_regex",
                         type="string",
                         help="Used with --gene-tag. "
                         "Ignore reads where the gene-tag matches this regex",
                         default="^(__|Unassigned)")

        group.add_option("--per-contig", dest="per_contig", action="store_true",
                         default=False,
                         help="group/dedup/count UMIs per contig (field 3 in BAM; RNAME),"
                         " e.g for transcriptome where contig = gene")

        group.add_option("--gene-transcript-map", dest="gene_transcript_map",
                         type="string",
                         help="File mapping transcripts to genes (tab separated)",
                         default=None)

        group.add_option("--per-cell", dest="per_cell", action="store_true",
                         default=False,
                         help="group/dedup/count per cell")

        parser.add_option_group(group)

    if add_group_dedup_options:

        group = OptionGroup(parser, "group/dedup options")

        group.add_option("--buffer-whole-contig", dest="whole_contig",
                         action="store_true", default=False,
                         help="Read whole contig before outputting bundles: "
                         "guarantees that no reads are missed, but increases "
                         "memory usage")

        group.add_option("--whole-contig", dest="whole_contig",
                         action="store_true", default=False,
                         help=optparse.SUPPRESS_HELP)

        group.add_option("--multimapping-detection-method",
                         dest="detection_method", type="choice",
                         choices=("NH", "X0", "XT"),
                         default=None,
                         help="Some aligners identify multimapping using bam "
                         "tags. Setting this option to NH, X0 or XT will "
                         "use these tags when selecting the best read "
                         "amongst reads with the same position and umi "
                         "[default=%default]")

        parser.add_option_group(group)

    # options added separately here to maintain better output order
    if add_sam_options:
        group = OptionGroup(parser, "SAM/BAM options")

        group.add_option("--mapping-quality", dest="mapping_quality",
                         type="int",
                         help="Minimum mapping quality for a read to be retained"
                         " [default=%default]",
                         default=0)

        group.add_option("--output-unmapped", dest="output_unmapped", action="store_true",
                         default=False, help=optparse.SUPPRESS_HELP)

        group.add_option("--unmapped-reads", dest="unmapped_reads",
                         type="choice",
                         choices=("discard", "use", "output"),
                         default="discard",
                         help=("How to handle unmapped reads. Options are "
                               "'discard', 'use' or 'correct' [default=%default]"))

        group.add_option("--chimeric-pairs", dest="chimeric_pairs",
                         type="choice",
                         choices=("discard", "use", "output"),
                         default="use",
                         help=("How to handle chimeric read pairs. Options are "
                               "'discard', 'use' or 'correct' [default=%default]"))

        group.add_option("--unpaired-reads", dest="unpaired_reads",
                         type="choice",
                         choices=("discard", "use", "output"),
                         default="use",
                         help=("How to handle unpaired reads. Options are "
                               "'discard', 'use' or 'correct' [default=%default]"))

        group.add_option("--ignore-umi", dest="ignore_umi",
                         action="store_true", help="Ignore UMI and dedup"
                         " only on position", default=False)

        group.add_option("--ignore-tlen", dest="ignore_tlen", action="store_true",
                         default=False,
                         help="Option to dedup paired end reads based solely on read1, "
                         "whether or not the template length is the same")

        group.add_option("--chrom", dest="chrom", type="string",
                         help="Restrict to one chromosome",
                         default=None)

        group.add_option("--subset", dest="subset", type="float",
                         help="Use only a fraction of reads, specified by subset",
                         default=None)

        group.add_option("-i", "--in-sam", dest="in_sam", action="store_true",
                         help="Input file is in sam format [default=%default]",
                         default=False)

        group.add_option("--paired", dest="paired", action="store_true",
                         default=False,
                         help="paired input BAM. [default=%default]")

        group.add_option("-o", "--out-sam", dest="out_sam", action="store_true",
                         help="Output alignments in sam format [default=%default]",
                         default=False)

        group.add_option("--no-sort-output", dest="no_sort_output",
                         action="store_true", default=False,
                         help="Don't Sort the output")

        parser.add_option_group(group)

    if add_pipe_options:
        group = OptionGroup(parser, "input/output options")
        group.add_option("-I", "--stdin", dest="stdin", type="string",
                         help="file to read stdin from [default = stdin].",
                         metavar="FILE")
        group.add_option("-L", "--log", dest="stdlog", type="string",
                         help="file with logging information "
                         "[default = stdout].",
                         metavar="FILE")
        group.add_option("-E", "--error", dest="stderr", type="string",
                         help="file with error information "
                         "[default = stderr].",
                         metavar="FILE")
        group.add_option("-S", "--stdout", dest="stdout", type="string",
                         help="file where output is to go "
                         "[default = stdout].",
                         metavar="FILE")
        group.add_option("--temp-dir", dest="tmpdir", type="string",
                         help="Directory for temporary files. If not set,"
                         " the bash environmental variable TMPDIR is used"
                         "[default = None].",
                         metavar="FILE")
        group.add_option("--log2stderr", dest="log2stderr",
                         action="store_true", help="send logging information"
                         " to stderr [default = False].")
        group.add_option("--compresslevel", dest="compresslevel", type="int",
                         help="Level of Gzip compression to use. Default (6) matches"
                         "GNU gzip rather than python gzip default (which is 9)")

        parser.set_defaults(stderr=sys.stderr)
        parser.set_defaults(stdout=sys.stdout)
        parser.set_defaults(stdlog=sys.stdout)
        parser.set_defaults(stdin=sys.stdin)
        parser.set_defaults(tmpdir=None)
        parser.set_defaults(log2stderr=False)
        parser.set_defaults(compresslevel=6)

    parser.add_option_group(group)

    group = OptionGroup(parser, "profiling options")

    group.add_option("--timeit", dest='timeit_file', type="string",
                     help="store timeing information in file [%default].")

    group.add_option("--timeit-name", dest='timeit_name', type="string",
                     help="name in timing file for this class of jobs "
                     "[%default].")

    group.add_option("--timeit-header", dest='timeit_header',
                     action="store_true",
                     help="add header for timing information [%default].")

    parser.add_option_group(group)

    group = OptionGroup(parser, "common options")

    group.add_option("-v", "--verbose", dest="loglevel", type="int",
                     help="loglevel [%default]. The higher, the more output.")

    group.add_option("-h", "--help", dest="short_help", action="callback",
                     callback=callbackShortHelp,
                     help="output short help (command line options only).")

    group.add_option('--help-extended', action='help',
                     help='Output full documentation')

    group.add_option("--random-seed", dest='random_seed', type="int",
                     help="random seed to initialize number generator "
                     "with [%default].")

    parser.add_option_group(group)

    # restore user defaults
    parser.defaults.update(user_defaults)

    if quiet:
        parser.set_defaults(loglevel=0)
    else:
        parser.set_defaults(loglevel=1)

    parser.set_defaults(
        timeit_file=None,
        timeit_name='all',
        timeit_header=None,
        random_seed=None,
    )

    if return_parser:
        return parser

    global_options, global_args = parser.parse_args(argv[1:])

    if global_options.random_seed is not None:
        random.seed(global_options.random_seed)

    if add_pipe_options:
        if global_options.stdout != sys.stdout:
            global_options.stdout = openFile(global_options.stdout, "w")
        if global_options.stderr != sys.stderr:
            if global_options.stderr == "stderr":
                global_options.stderr = global_options.stderr
            else:
                global_options.stderr = openFile(global_options.stderr, "w")
        if global_options.stdlog != sys.stdout:
            global_options.stdlog = openFile(global_options.stdlog, "a")
        elif global_options.log2stderr:
            global_options.stdlog = global_options.stderr
        if global_options.stdin != sys.stdin:
            global_options.stdin = openFile(global_options.stdin, "r")
    else:
        global_options.stderr = sys.stderr
        global_options.stdout = sys.stdout
        global_options.stdin = sys.stdin
        if global_options.log2stderr:
            global_options.stdlog = sys.stderr
        else:
            global_options.stdlog = sys.stdout

    if global_options.loglevel >= 1:
        global_options.stdlog.write(getHeader() + "\n")
        global_options.stdlog.write(getParams(global_options) + "\n")
        global_options.stdlog.flush()

    # configure logging
    # map from 0-10 to logging scale
    # 0: quiet
    # 1: little verbositiy
    # >1: increased verbosity
    if global_options.loglevel == 0:
        lvl = logging.ERROR
    elif global_options.loglevel == 1:
        lvl = logging.INFO
    else:
        lvl = logging.DEBUG

    if global_options.stdout == global_options.stdlog:
        format = '# %(asctime)s %(levelname)s %(message)s'
    else:
        format = '%(asctime)s %(levelname)s %(message)s'

    logging.basicConfig(
        level=lvl,
        format=format,
        stream=global_options.stdlog)

    # set up multi-line logging
    # Note that .handlers is not part of the API, might change
    # Solution is to configure handlers explicitely.
    for handler in logging.getLogger().handlers:
        handler.setFormatter(MultiLineFormatter(format))

    return global_options, global_args


def validateExtractOptions(options):
    ''' Check the validity of the option combinations for barcode extraction'''

    if not options.pattern and not options.pattern2:
        if not options.read2_in:
            raise ValueError("Must supply --bc-pattern for single-end")
        else:
            raise ValueError("Must supply --bc-pattern and/or --bc-pattern2 "
                             "if paired-end ")

    if options.pattern2:
        if not options.read2_in:
            raise ValueError("must specify a paired fastq ``--read2-in``")

        if not options.pattern2:
            options.pattern2 = options.pattern

    if options.filtered_out2 and not options.read2_in:
        raise ValueError("Cannot use --filtered-out2 without read2 input (--read2-in)")

    if ((options.read2_in and options.filtered_out) and not options.filtered_out2) or (
            options.filtered_out2 and not options.filtered_out):
        raise ValueError("Must supply both --filtered-out and --filtered-out2"
                         "to write out filtered reads for paired end")

    extract_cell = False
    extract_umi = False

    # If the pattern is a regex we can compile the regex(es) prior to
    # ExtractFilterAndUpdate instantiation
    if options.extract_method == "regex":
        if options.pattern:
            try:
                options.pattern = regex.compile(options.pattern)
            except regex.error:
                raise ValueError("--bc-pattern '%s' is not a "
                                 "valid regex" % options.pattern)

        if options.pattern2:
            try:
                options.pattern2 = regex.compile(options.pattern2)
            except regex.Error:
                raise ValueError("--bc-pattern2 '%s' is not a "
                                 "valid regex" % options.pattern2)

    # check whether the regex contains a umi group(s) and cell groups(s)
    if options.extract_method == "regex":
        if options.pattern:
            for group in options.pattern.groupindex:
                if group.startswith("cell_"):
                    extract_cell = True
                elif group.startswith("umi_"):
                    extract_umi = True
        if options.pattern2:
            for group in options.pattern2.groupindex:
                if group.startswith("cell_"):
                    extract_cell = True
                elif group.startswith("umi_"):
                    extract_umi = True

    # check whether the pattern string contains umi/cell bases
    elif options.extract_method == "string":
        if options.pattern:
            if "C" in options.pattern:
                extract_cell = True
            if "N" in options.pattern:
                extract_umi = True
        if options.pattern2:
            if "C" in options.pattern2:
                extract_cell = True
            if "N" in options.pattern2:
                extract_umi = True

    if not extract_umi:
        if options.extract_method == "string":
            raise ValueError("barcode pattern(s) do not include any umi bases "
                             "(marked with 'Ns') %s, %s" % (
                                 options.pattern, options.pattern2))
        elif options.extract_method == "regex":
            raise ValueError("barcode regex(es) do not include any umi groups "
                             "(starting with 'umi_') %s, %s" % (
                                 options.pattern, options.pattern2))

    return(extract_cell, extract_umi)


def validateSamOptions(options, group=False):
    ''' Check the validity of the option combinations for sam/bam input '''

    if options.per_gene:
        if options.gene_tag and options.per_contig:
            raise ValueError("need to use either --per-contig "
                             "OR --gene-tag, please do not provide both")

        if not options.per_contig and not options.gene_tag:
            raise ValueError("for per-gene applications, must supply "
                             "--per-contig or --gene-tag")

    if options.per_contig and not options.per_gene:
        raise ValueError("need to use --per-gene with --per-contig")

    if options.gene_tag and not options.per_gene:
        raise ValueError("need to use --per-gene with --gene_tag")

    if options.gene_transcript_map and not options.per_contig:
        raise ValueError("need to use --per-contig and --per-gene"
                         "with --gene-transcript-map")

    if options.get_umi_method == "tag":
        if options.umi_tag is None:
            raise ValueError("Need to supply the --umi-tag option")
        if options.per_cell and options.cell_tag is None:
            raise ValueError("Need to supply the --cell-tag option")

    if options.assigned_tag is None:
        options.assigned_tag = options.gene_tag

    if options.skip_regex:
        try:
            re.compile(options.skip_regex)
        except re.error:
            raise ValueError("skip-regex '%s' is not a "
                             "valid regex" % options.skip_regex)

    if not group:
        if options.unmapped_reads == "output":
            raise ValueError("Cannot use --unmapped-reads=output. If you want "
                             "to retain unmapped without deduplicating them, "
                             "use the group command")
        if options.chimeric_pairs == "output":
            raise ValueError("Cannot use --chimeric-pairs=output. If you want "
                             "to retain chimeric read pairs without "
                             "deduplicating them, use the group command")
        if options.unpaired_reads == "output":
            raise ValueError("Cannot use --unpaired-reads=output. If you want "
                             "to retain unmapped without deduplicating them, "
                             "use the group command")

    if options.paired:
        if options.chimeric_pairs == "use":
            warn("Chimeric read pairs are being used. "
                 "Some read pair UMIs may be grouped/deduplicated using "
                 "just the mapping coordinates from read1."
                 "This may also increase the run time and memory usage. "
                 "Consider --chimeric-pairs==discard to discard these reads "
                 "or --chimeric-pairs==output (group command only) to "
                 "output them without grouping")
        if options.unpaired_reads == "use":
            warn("Unpaired read pairs are being used. "
                 "Some read pair UMIs may be grouped/deduplicated using "
                 "just the mapping coordinates from read1."
                 "This may also increase the run time and memory usage. "
                 "Consider --unpared-reads==discard to discard these reads "
                 "or --unpared-reads==output (group command only) to "
                 "output them without grouping")
        if options.unmapped_reads == "use":
            warn("Unmapped read pairs are being used. "
                 "Some read pair UMIs may be grouped/deduplicated using "
                 "just the mapping coordinates from read1. "
                 "This may also increase the run time and memory usage. "
                 "Consider --unmapped_reads==discard to discard these reads "
                 "or --unmapped_reads==output (group command only) to "
                 "output them without grouping")

    command = " ".join(sys.argv)
    info("command: %s" % command)
    if "--umi-tag" in command or "--cell-tag" in command:
        if options.get_umi_method != "tag":
            raise ValueError("--umi-tag and/or --cell-tag options provided. "
                             "Need to set --extract-umi-method=tag")

    if options.unmapped_reads == "use":
        if not options.paired:
            raise ValueError("--unmapped-reads=use is only compatible with "
                             "paired end reads (--paired)")

    if "--chimeric-pairs" in command:
        info("command: %s" % command)
        if not options.paired:
            raise ValueError("--chimeric-pairs is only compatible "
                             "with paired end reads (--paired)")

    if "--unpaired-reads" in command:
        if not options.paired:
            raise ValueError("--unpaired-reads is only compatible "
                             "with paired end reads (--paired)")

    if "--ignore-tlen" in command:
        if not options.paired:
            raise ValueError("--ignore-tlen is only compatible "
                             "with paired end reads (--paired)")

    # legacy support for --output-unmapped behaviour
    if options.output_unmapped:
        warn("--output-unmapped will be removed in the near future. "
             "Use --unmapped-reads=output instead")
        # We will update the value of options.unmapped_reads so we want to
        # check the user has not also supplied this option
        if "--unmapped_reads" in command:
            raise ValueError("Do not use --output-unmapped in combination with"
                             "--unmapped-reads. Just use --unmapped-reads")
        options.unmapped_reads = "output"


def Stop():
    """stop the experiment.

    This method performs final book-keeping, closes the output streams
    and writes the final log messages indicating script completion.
    """

    if global_options.loglevel >= 1 and global_benchmark:
        t = time.time() - global_starting_time
        global_options.stdlog.write(
            "######### Time spent in benchmarked functions #########\n")
        global_options.stdlog.write("# function\tseconds\tpercent\n")
        for key, value in global_benchmark.items():
            global_options.stdlog.write(
                "# %s\t%6i\t%5.2f%%\n" % (key, value,
                                          (100.0 * float(value) / t)))
        global_options.stdlog.write(
            "#######################################################\n")

    if global_options.loglevel >= 1:
        global_options.stdlog.write(getFooter() + "\n")

    # close files
    if global_options.stdout != sys.stdout:
        global_options.stdout.close()
    # do not close log, otherwise error occurs in atext.py
    # if global_options.stdlog != sys.stdout:
    #   global_options.stdlog.close()

    if global_options.stderr != sys.stderr:
        global_options.stderr.close()

    if global_options.timeit_file:

        outfile = open(global_options.timeit_file, "a")

        if global_options.timeit_header:
            outfile.write("\t".join(
                ("name", "wall", "user", "sys", "cuser", "csys",
                 "host", "system", "release", "machine",
                 "start", "end", "path", "cmd")) + "\n")

        csystem, host, release, version, machine = map(str, os.uname())
        uusr, usys, c_usr, c_sys = map(lambda x: "%5.2f" % x, os.times()[:4])
        t_end = time.time()
        c_wall = "%5.2f" % (t_end - global_starting_time)

        if sys.argv[0] == "run.py":
            cmd = global_args[0]
            if len(global_args) > 1:
                cmd += " '" + "' '".join(global_args[1:]) + "'"
        else:
            cmd = sys.argv[0]

        result = "\t".join((global_options.timeit_name,
                            c_wall, uusr, usys, c_usr, c_sys,
                            host, csystem, release, machine,
                            time.asctime(time.localtime(global_starting_time)),
                            time.asctime(time.localtime(t_end)),
                            os.path.abspath(os.getcwd()),
                            cmd)) + "\n"

        outfile.write(result)
        outfile.close()


def log(loglevel, message):
    """log message at loglevel."""
    logging.log(loglevel, message)


def info(message):
    '''log information message, see the :mod:`logging` module'''
    logging.info(message)


def warning(message):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning(message)


def warn(message):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning(message)


def debug(message):
    '''log debugging message, see the :mod:`logging` module'''
    logging.debug(message)


def error(message):
    '''log error message, see the :mod:`logging` module'''
    logging.error(message)
    raise ValueError("UMI-tools failed with an error. Check the log file")


def critical(message):
    '''log critical message, see the :mod:`logging` module'''
    logging.critical(message)


def getTempFile(dir=None, shared=False, suffix=""):
    '''get a temporary file.

    The file is created and the caller needs to close and delete
    the temporary file once it is not used any more.

    Arguments
    ---------
    dir : string
        Directory of the temporary file and if not given is set to the
        default temporary location in the global configuration dictionary.
    shared : bool
        If set, the tempory file will be in a shared temporary
        location (given by the global configuration directory).
    suffix : string
        Filename suffix

    Returns
    -------
    file : File
        A file object of the temporary file.
    '''

    return tempfile.NamedTemporaryFile(dir=dir, delete=False, prefix="ctmp",
                                       suffix=suffix)


def getTempFilename(dir=None, shared=False, suffix=""):
    '''return a temporary filename.

    The file is created and the caller needs to delete the temporary
    file once it is not used any more.

    Arguments
    ---------
    dir : string
        Directory of the temporary file and if not given is set to the
        default temporary location in the global configuration dictionary.
    shared : bool
        If set, the tempory file will be in a shared temporary
        location.
    suffix : string
        Filename suffix

    Returns
    -------
    filename : string
        Absolute pathname of temporary file.

    '''
    tmpfile = getTempFile(dir=dir, shared=shared, suffix=suffix)
    tmpfile.close()
    return tmpfile.name










def getUserDefinedBarcodes(whitelist_tsv, whitelist_tsv2=None,
                           getErrorCorrection=False,
                           deriveErrorCorrection=False,
                           threshold=1):
    '''
    whitelist_tsv: tab-separated file with whitelisted barcodes. First
    field should be whitelist barcodes. Second field [optional] should
    be comma-separated barcodes which are to be corrected to the
    barcode in the first field.

    whitelist_tsv2: as above but for read2s
    getErrorCorrection: extract the second field in whitelist_tsv and
    return a map of non-whitelist:whitelist

    deriveErrorCorrection: return a map of non-whitelist:whitelist
    using a simple edit distance threshold
    '''

    base2errors = {"A": ["T", "C", "G", "N"],
                   "T": ["A", "C", "G", "N"],
                   "C": ["T", "A", "G", "N"],
                   "G": ["T", "C", "A", "N"]}

    whitelist = []

    if getErrorCorrection or deriveErrorCorrection:
        false_to_true_map = {}
    else:
        false_to_true_map = None

    def singleBarcodeGenerator(whitelist_tsv):
        with openFile(whitelist_tsv, "r") as inf:
            for line in inf:
                if line.startswith('#'):
                    continue
                line = line.strip().split("\t")
                yield(line[0])

    def pairedBarcodeGenerator(whitelist_tsv, whitelist_tsv2):

        whitelist1 = []
        whitelist2 = []

        with openFile(whitelist_tsv, "r") as inf:
            for line in inf:
                if line.startswith('#'):
                    continue

                line = line.strip().split("\t")
                whitelist1.append(line[0])

        with openFile(whitelist_tsv2, "r") as inf2:
            for line in inf2:
                if line.startswith('#'):
                    continue

                line = line.strip().split("\t")
                whitelist2.append(line[0])

        for w1, w2 in itertools.product(whitelist1, whitelist2):
            yield(w1 + w2)

    if deriveErrorCorrection:

        if whitelist_tsv2:
            whitelist_barcodes = pairedBarcodeGenerator(whitelist_tsv, whitelist_tsv2)
        else:
            whitelist_barcodes = singleBarcodeGenerator(whitelist_tsv)

        for whitelist_barcode in whitelist_barcodes:
            whitelist.append(whitelist_barcode)

            # for every possible combination of positions for error(s)
            for positions in itertools.product(
                    range(0, len(whitelist_barcode)), repeat=threshold):

                m_bases = [base2errors[whitelist_barcode[x]] for x in positions]

                # for every possible combination of errors
                for m in itertools.product(*m_bases):
                    error_barcode = list(whitelist_barcode)

                    # add errors
                    for pos, error_base in zip(positions, m):
                        error_barcode[pos] = error_base

                    error_barcode = "".join(error_barcode)

                    # if error barcode has already been seen, must be within
                    # threshold edit distance of >1 whitelisted barcodes
                    if error_barcode in false_to_true_map:
                        # don't report multiple times for the same barcode
                        if false_to_true_map[error_barcode]:
                            info("Error barcode %s can be assigned to more than "
                                   "one possible true barcode: %s or %s" % (
                                       error_barcode,
                                       false_to_true_map[error_barcode],
                                       whitelist_barcode))
                        false_to_true_map[error_barcode] = None
                    else:
                        false_to_true_map[error_barcode] = whitelist_barcode

    elif getErrorCorrection:
        assert not whitelist_tsv2, ("Can only extract errors from the whitelist "
                                    "if a single whitelist is given")
        with openFile(whitelist_tsv, "r") as inf:

            for line in inf:

                if line.startswith('#'):
                    continue

                line = line.strip().split("\t")
                whitelist_barcode = line[0]
                whitelist.append(whitelist_barcode)

                if getErrorCorrection:
                    for error_barcode in line[1].split(","):
                        false_to_true_map[error_barcode] = whitelist_barcode

    else:  # no error correction
        if whitelist_tsv2:
            whitelist_barcodes = pairedBarcodeGenerator(whitelist_tsv, whitelist_tsv2)
        else:
            whitelist_barcodes = singleBarcodeGenerator(whitelist_tsv)

        whitelist = [x for x in whitelist_barcodes]

    return set(whitelist), false_to_true_map


import os, requests
from ftplib import FTP
from tqdm import tqdm

def lower_list(input: list): return [v.lower() for v in input]

def lower_dict(input: dict): 
    return dict((k.lower(), v.lower) for k, v in input.items())

class SplitFastq:
    def __init__(
        self,
        file:str,
        n_split:int,
        out_name:str,
        out_path:str='./',
        silence:bool=False,
        ):
        """fastq file을 원하는 수 만큼 균등하게 나눠주는 함수.

        Args:
            file (str): fastq 파일 경로
            n_split (int): 몇 등분 할 것인지 적는 칸
            out_name (str): 나눴을 때 저장되는 파일들의 prefix
            out_path (str, optional): Output이 저장 될 경로. Defaults to './'.
            silence (bool, optional): Logging을 위한 print 되는 메시지를 끄는 용도. Defaults to False.
        """        
        
        output_format = 'fastq'
        lineset = 4

        self.names = []
        self.dir   = '%s/%s_subsplits' % (os.path.abspath(out_path), out_name)
        os.makedirs(self.dir, exist_ok=True)

        with open(file, 'r') as f:
            lines   = f.readlines()
            total   = len(lines)
            rec_cnt = total / lineset

            list_nBins = [[int(rec_cnt * (i + 0) / n_split), int(rec_cnt * (i + 1) / n_split)] for i in range(n_split)]
            self.meta  = {}
            cnt = 0

            for nStart, nEnd in list_nBins:
                if silence == False: print('[Info] Make data subsplits: %s - %s' % (nStart, nEnd))

                sSplit_file_name = '%s_%s.%s' % (out_name, cnt, output_format)
                with open('%s/%s' % (self.dir, sSplit_file_name), 'w') as outfile:
                    for l in lines[nStart*lineset:nEnd*lineset]: outfile.write(l)
                
                self.names.append(sSplit_file_name)
                
                
                self.meta[sSplit_file_name] = {
                    'start': nStart,
                    'end'  : nEnd,
                    'count': nEnd - nStart
                }
                cnt += 1

# class END: SplitFastq

def request_file(server:str, remote_path:str, local_path:str, target_file:str, silence:bool=False):
    
    url  = f"https://{server}/{remote_path}/{target_file}"
    save = f"{local_path}/{target_file}"
    
    os.makedirs(local_path, exist_ok=True)
    
    # FTP Connect
    with FTP(server) as ftp:
        ftp.login()

        # Get file size
        file_size = ftp.size(f'{remote_path}/{target_file}')

    with requests.get(url, stream=True) as response, open(save, "wb") as file, tqdm(
        total=file_size,
        desc=f"[Info] Downloading {target_file}",
        unit="B",
        unit_scale=True,
        unit_divisor=1024,
    ) as pbar:
        
        for data in response.iter_content(chunk_size=4096):
            file.write(data)
            pbar.update(len(data))

    if silence==False: print(f"[Info] File downloaded successfully: {save}")
    
# def End: request_file

def download_file_ftp(server:str, remote_path:str, local_path:str, target_file:str, user_name:str=None, user_pwd:str=None):
    '''Download specific file from FTP server.
    하지만 https를 통해서 다운로드 받는 것이 속도나 안정성에서 더 나은듯...
    request로 전부 대체될 예정.
    
    ### Example
    ```python
    import os
    from genet.utils import download_file_ftp

    ftp_server  = "ftp.ncbi.nlm.nih.gov"
    ftp_user    = "your_username"
    remote_path = "/genomes/ASSEMBLY_REPORTS/"
    local_path  = "genet/database/metadata/NCBI/"
    target_file = 'assembly_summary_refseq.txt'

    download_file_ftp(ftp_server, remote_path, local_path, target_file, ftp_user)
    ```
    '''
        
    remote_filepath = f'{remote_path}/{target_file}'
    local_filepath  = f'{local_path}/{target_file}'
    
    os.makedirs(local_path, exist_ok=True)

    try:
        # FTP Connect
        with FTP(server) as ftp:
            ftp.login()

            # Get file size
            file_size = ftp.size(remote_filepath)

            # Progress bar 
            with open(local_filepath, "wb") as local_file, tqdm(
                    desc=f"[Info] Downloading {target_file}",
                    total=file_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
            ) as pbar:
                
                def callback(data):
                    local_file.write(data)
                    pbar.update(len(data))

                ftp.retrbinary(f"RETR {remote_filepath}", callback)
            
    except Exception as e:
        print(f"Error: {e}")




